"""
A module containing unit tests for the `wcsutil` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
from itertools import groupby

import pytest

from tweakwcs.utils.jwst_utils import assign_jwst_tweakwcs_groups


@pytest.fixture(scope='function')
def data_model_list():
    pytest.importorskip("jwst")
    from jwst.datamodels import DataModel, ModelContainer

    models = []
    for k in range(6):
        m = DataModel()
        m.meta.observation.program_number = '0001'
        m.meta.observation.observation_number = '1'
        m.meta.observation.visit_number = '1'
        m.meta.observation.visit_group = '1'
        m.meta.observation.sequence_id = '01'
        m.meta.observation.activity_id = '1'
        m.meta.observation.exposure_number = '1'
        m.meta.instrument.name = 'NIRCAM'
        m.meta.instrument.channel = 'SHORT'
        m.meta.filename = 'file{:d}.fits'.format(k)
        models.append(m)

    models[-3].meta.observation.observation_number = '2'
    models[-2].meta.observation.observation_number = '3'
    models[-1].meta.observation.observation_number = '3'

    return ModelContainer(models)


@pytest.fixture
def defective_data_model():
    class DummyMeta(dict):
        def __init__(self, *args):
            super().__init__(args)
            dict.__setitem__(self, 'filename', 'dummy.file')

        @property
        def tweakwcs_group_id(self):
            return dict.__getitem__(self, 'tweakwcs_group_id')

        @tweakwcs_group_id.setter
        def tweakwcs_group_id(self, v):
            dict.__setitem__(self, 'tweakwcs_group_id', v)

    class BrokenModel():
        def __init__(self):
            self._meta = DummyMeta()

        @property
        def meta(self):
            return self._meta

        def extend_schema(self, *args, **kwargs):
            pass

    return BrokenModel()


def test_assign_jwst_tweakwcs_groups_fail(defective_data_model):
    pytest.importorskip("jwst")
    assign_jwst_tweakwcs_groups([defective_data_model])
    assert defective_data_model.meta.tweakwcs_group_id == 'None'


def test_assign_jwst_tweakwcs_groups(data_model_list, monkeypatch):
    """ Imitate string file names and file IO: """
    jwst = pytest.importorskip("jwst")
    models = {m.meta.filename: m for m in data_model_list}
    monkeypatch.setattr(jwst.datamodels.DataModel, 'save', lambda s, m: m)
    monkeypatch.setattr(jwst.datamodels, 'open', lambda fname: models[fname])

    assign_jwst_tweakwcs_groups(list(models.keys()))

    gids = sorted([model.meta.tweakwcs_group_id for model in data_model_list])
    assert sorted([len(list(g)) for k, g in groupby(gids)]) == [1, 2, 3]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A module that provides algorithms for performing linear fit between
sets of 2D points.

:Authors: Mihai Cara, Warren Hack

:License: :doc:`../LICENSE`

"""
import logging
import numbers
import numpy as np
from astropy.modeling.fitting import LevMarLSQFitter, _fitter_to_model_params

from .linalg import inv
from . import __version__  # noqa: F401

__author__ = 'Mihai Cara, Warren Hack'

__all__ = [
    'iter_linear_fit', 'build_fit_matrix', 'SUPPORTED_FITGEOM_MODES',
    '_LevMarLSQFitter2x2'
]

# Supported fitgeom modes and corresponding minobj
SUPPORTED_FITGEOM_MODES = {
    'shift': 1,
    'rshift': 2,
    'rscale': 2,
    'general': 3
}

_FITGEOM_KEYS = tuple(SUPPORTED_FITGEOM_MODES.keys())
_SUPPORTED_FITGEOM_EN_STR = (', '.join(map(repr, _FITGEOM_KEYS[:-1])) +
                             ', or ' + repr(_FITGEOM_KEYS[-1]))


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class SingularMatrixError(Exception):
    """ An error class used to report when a singular matrix is encountered."""
    pass


class NotEnoughPointsError(Exception):
    """
    An error class used to report when there are not enough points to
    find parameters of a linear transformation.
    """
    pass


def iter_linear_fit(xy, uv, wxy=None, wuv=None,
                    fitgeom='general', center=None,
                    nclip=3, sigma=(3.0, 'rmse'), clip_accum=False):
    r"""
    Compute linear transformation parameters that "best" (in the sense of
    minimizing residuals) transform ``uv`` source position to ``xy``
    sources iteratively using sigma-clipping.

    More precisely, this functions attempts to find a ``2x2`` matrix ``F`` and
    a shift vector ``s`` that minimize the residuals between the *transformed*
    reference source coordinates ``uv``

    .. math::
        \mathbf{xy}'_k = \mathbf{F}\cdot(\mathbf{uv}_k-\mathbf{c})+\
        \mathbf{s} + \mathbf{c}
        :label: ilf1

    and the "observed" source positions ``xy``:

    .. math::
        \epsilon^2 = \Sigma_k w_k \|\mathbf{xy}_k-\mathbf{xy}'_k\|^2.
        :label: ilf2

    In the above equations, :math:`\mathbf{F}` is a ``2x2`` matrix while
    :math:`\mathbf{xy}_k` and :math:`\mathbf{uv}_k` are the position
    coordinates of the ``k``-th source (row in input ``xy`` and ``uv`` arrays).

    One of the two catalogs (``xy`` or ``uv``) contains what we refer to as
    "image" source positions and the other one as "reference" source positions.
    The meaning assigned to ``xy`` and ``uv`` parameters are up to the
    caller of this function.

    Parameters
    ----------
    xy: numpy.ndarray
        A ``(N, 2)``-shaped array of source positions (one 2-coordinate
        position per line).

    uv: numpy.ndarray
        A ``(N, 2)``-shaped array of source positions (one 2-coordinate
        position per line). This array *must have* the same length (shape)
        as the ``xy`` array.

    wxy: numpy.ndarray, None, optional
        A 1-dimensional array of weights of the same length (``N``)
        as ``xy`` array indicating how much a given coordinate should be
        weighted in the fit. If not provided or set to `None`, all positions
        will be contribute equally to the fit if ``wuv`` is also set to `None`.
        See ``Notes`` section for more details.

    wuv: numpy.ndarray, None, optional
        A 1-dimensional array of weights of the same length (``N``)
        as ``xy`` array indicating how much a given coordinate should be
        weighted in the fit. If not provided or set to `None`, all positions
        will be contribute equally to the fit if ``wxy`` is also set to `None`.
        See ``Notes`` section for more details.

    fitgeom: {'shift', 'rshift', 'rscale', 'general'}, optional
        The fitting geometry to be used in fitting the matched object lists.
        This parameter is used in fitting the shifts (offsets), rotations
        and/or scale changes from the matched object lists. The 'general'
        fit geometry allows for independent scale and rotation for each axis.

    center: tuple, list, numpy.ndarray, None, optional
        A list-like container with two ``X``- and ``Y``-positions of the center
        (origin) of rotations in the ``uv`` and ``xy`` coordinate frames.
        If not provided, ``center`` is estimated as a (weighted) mean position
        in the ``uv`` frame.

    nclip: int, None, optional
        Number (a non-negative integer) of clipping iterations in fit.
        Clipping will be turned off if ``nclip`` is either `None` or 0.

    sigma: float, tuple of the form (float, str), optional
        When a tuple is provided, first value (a positive number)
        indicates the number of "fit error estimates" to use for clipping.
        The second value (a string) indicates the statistic to be
        used for "fit error estimate". Currently the following values are
        supported: ``'rmse'``, ``'mae'``, and ``'std'``
        - see ``Notes`` section for more details.

        When ``sigma`` is a single number, it must be a positive number and
        the default error estimate ``'rmse'`` is assumed.

        This parameter is ignored when ``nclip`` is either `None` or 0.

    clip_accum: bool, optional
        Indicates whether or not to reset the list of "bad" (clipped out)
        sources after each clipping iteration. When set to `True` the list
        only grows with each iteration as "bad" positions never re-enter the
        pool of available position for the fit. By default the list of
        "bad" source positions is purged at each iteration.

    Returns
    -------
    fit: dict
        - ``'shift'``: A ``numpy.ndarray`` with two components of the
          computed shift.
        - ``'shift_ld'``: A ``numpy.ndarray`` with two components of the
          computed shift of type ``numpy.longdouble``.
        - ``'matrix'``: A ``2x2`` ``numpy.ndarray`` with the computed
          generalized rotation matrix.
        - ``'matrix_ld'``: A ``2x2`` ``numpy.ndarray`` with the computed
          generalized rotation matrix of type ``numpy.longdouble``.
        - ``'proper_rot'``: Rotation angle (degree) as if the rotation is
          proper.
        - ``'rot'``: A tuple of ``(rotx, roty)`` - the rotation angles with
          regard to the ``X`` and ``Y`` axes.
        - ``'<rot>'``: *Arithmetic mean* of the angles of rotation around
          ``X`` and ``Y`` axes.
        - ``'scale'``: A tuple of ``(sx, sy)`` - scale change in the direction
          of the ``X`` and ``Y`` axes.
        - ``'<scale>'``: *Geometric mean* of scales ``sx`` and ``sy``.
        - ``'skew'``: Computed skew.
        - ``'proper'``: a boolean indicating whether the rotation is proper.
        - ``'fitgeom'``: Fit geometry (allowed transformations) used for
          fitting data (to minimize residuals). This is copy of the input
          argument ``fitgeom``.
        - ``'center'``: Center of rotation
        - ``'center_ld'``: Center of rotation as a ``numpy.longdouble``.
        - ``'fitmask'``: A boolean array indicating which source positions
          where used for fitting (`True`) and which were clipped out
          (`False`). **NOTE** For weighted fits, positions with zero
          weights are automatically excluded from the fits.
        - ``'eff_nclip'``: Effective number of clipping iterations
        - ``'rmse'``: Root-Mean-Square Error
        - ``'mae'``: Mean Absolute Error
        - ``'std'``: Standard Deviation of the residuals
        - ``'resids'``: An array of residuals of the fit.
          **NOTE:** Only the residuals for the "valid" points are reported
          here. Therefore the length of this array may be smaller than the
          length of input arrays of positions.

    Notes
    -----
    **Weights**

    Weights can be provided for both "image" source positions and "reference"
    source positions. When no weights are given, all positions are weighted
    equally. When only one set of positions have weights (i.e., either ``wxy``
    or ``wuv`` is not `None`) then weights in :eq:`ilf2` are set to be equal
    to the provided set of weights. When weights for *both* "image" source
    positions and "reference" source positions are provided, then the
    combined weight that is used in :eq:`ilf2` is computed as:

    .. math::
        1/w = 1/w_{xy} + 1/w_{uv}.

    **Statistics for clipping**

    Several statistics are available for clipping iterations and all of them
    are reported in the returned ``fit`` dictionary regardless of the
    setting in ``sigma``:

    .. math::
        \mathrm{RMSE} = \sqrt{\Sigma_k w_k \|\mathbf{r}_k\|^2}

    .. math::
        \mathrm{MAE} = \sqrt{\Sigma_k w_k \|\mathbf{r}_k\|}

    .. math::
        \mathrm{STD} = \sqrt{\Sigma_k w_k \|\mathbf{r}_k - \
                       \mathbf{\overline{r}}\|^2}/(1-V_2)

    where :math:`\mathbf{r}_k=\mathbf{xy}_k-\mathbf{xy}'_k`,
    :math:`\Sigma_k w_k = 1`, and :math:`V_2=\Sigma_k w_k^2`.

    """
    try:
        minobj = SUPPORTED_FITGEOM_MODES[fitgeom]

        if fitgeom == 'general':
            linear_fit = fit_general
        elif fitgeom == 'rscale':
            linear_fit = fit_rscale
        elif fitgeom == 'rshift':
            linear_fit = fit_rshift
        elif fitgeom == 'shift':
            linear_fit = fit_shifts
        else:  # pragma: no cover
            raise AssertionError("Notify developer: fitter selector block is "
                                 "out-of-sync with SUPPORTED_FITGEOM_MODES.")

    except KeyError:
        raise ValueError("Unsupported 'fitgeom' value: '{}'".format(fitgeom))

    xy = np.array(xy, dtype=np.longdouble)
    uv = np.array(uv, dtype=np.longdouble)

    if len(xy.shape) != 2 or xy.shape[1] != 2 or uv.shape != xy.shape:
        raise ValueError("Input coordinate arrays 'xy' and 'uv' must be of "
                         "shape (N, 2) where N is the number of coordinate "
                         "points.")

    wmask = np.ones(len(xy), dtype=np.bool_)

    if wxy is not None:
        wxy = np.asarray(wxy)
        if len(wxy.shape) != 1 or wxy.shape[0] != xy.shape[0]:
            raise ValueError("Weights 'wxy' must be a 1-dimensional vector "
                             "of lengths equal to the number of input points.")
        wmask *= wxy > 0.0

    if wuv is not None:
        wuv = np.asarray(wuv)
        if len(wuv.shape) != 1 or wuv.shape[0] != xy.shape[0]:
            raise ValueError("Weights 'wuv' must be a 1-dimensional vector "
                             "of lengths equal to the number of input points.")
        wmask *= wuv > 0.0

    mask = wmask

    if sigma is None and nclip is not None and nclip > 0:
        raise ValueError("Argument 'sigma' cannot be None when 'nclip' is "
                         "a positive number.")

    if isinstance(sigma, numbers.Number):
        sigstat = 'rmse'  # default value
        nsigma = float(sigma)

    elif sigma is not None:
        nsigma = float(sigma[0])
        sigstat = sigma[1]
        if sigstat not in ['rmse', 'mae', 'std']:
            raise ValueError("Unsupported sigma statistics value.")

    if sigma is not None and nsigma <= 0.0:
        raise ValueError("The value of sigma for clipping iterations must be "
                         "positive.")

    if nclip is None:
        nclip = 0
    else:
        if nclip < 0:
            raise ValueError("Argument 'nclip' must be non-negative.")
        nclip = int(nclip)

    if np.count_nonzero(mask) == minobj:
        log.warning("The number of sources for the fit is smaller than the "
                    "minimum number of sources necessary for the requested "
                    "'fitgeom'.")
        log.warning("Resetting number of clipping iterations to 0.")
        nclip = 0

    if center is None:
        center_ld = uv[mask].mean(axis=0, dtype=np.longdouble)
        center = center_ld.astype(np.double)
    else:
        center_ld = np.longdouble(center)

    xy[mask] -= center_ld
    uv[mask] -= center_ld

    log.info("Performing '{:s}' fit".format(fitgeom))

    # initial fit:
    wmxy = None if wxy is None else wxy[mask]
    wmuv = None if wuv is None else wuv[mask]
    fit = linear_fit(xy[mask], uv[mask], wmxy, wmuv)

    # clipping iterations:
    effective_nclip = 0
    for n in range(nclip):
        resids = fit['resids']

        # redefine what pixels will be included in next iteration
        cutoff = nsigma * fit[sigstat]

        nonclipped = np.linalg.norm(resids, axis=1) < cutoff
        if np.count_nonzero(nonclipped) < minobj or nonclipped.all():
            break

        effective_nclip += 1

        prev_mask = mask
        if not clip_accum:
            mask = np.array(wmask)
        mask[prev_mask] *= nonclipped

        wmxy = None if wxy is None else wxy[mask]
        wmuv = None if wuv is None else wuv[mask]
        fit = linear_fit(xy[mask], uv[mask], wmxy, wmuv)

    fit['center'] = center
    fit['center_ld'] = center_ld
    fit['fitmask'] = mask
    fit['eff_nclip'] = effective_nclip
    return fit


def _compute_stat(fit, residuals, weights):
    if weights is None:
        fit['rmse'] = float(np.sqrt(np.mean(2 * residuals**2)))
        fit['mae'] = float(np.mean(np.linalg.norm(residuals, axis=1)))
        fit['std'] = float(np.linalg.norm(residuals.std(axis=0)))
    else:
        # assume all weights > 0 (this should be insured by the caller => no
        # need to repeat the check here)
        npts = len(weights)
        wt = np.sum(weights)
        if npts == 0 or wt == 0.0:
            fit['rmse'] = float('nan')
            fit['mae'] = float('nan')
            fit['std'] = float('nan')
            return

        w = weights / wt
        fit['rmse'] = float(np.sqrt(np.sum(np.dot(w, residuals**2))))
        fit['mae'] = float(np.dot(w, np.linalg.norm(residuals, axis=1)))

        if npts == 1:
            fit['std'] = 0.0
        else:
            # see:
            # https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Reliability_weights_2
            wmean = np.dot(w, residuals)
            fit['std'] = float(
                np.sqrt(np.sum(np.dot(w, (residuals - wmean)**2) /
                               (1.0 - np.sum(w**2))))
            )


def fit_shifts(xy, uv, wxy=None, wuv=None):
    """ Fits (non-iteratively and without sigma-clipping) a displacement
    transformation only between input lists of positions ``xy`` and ``uv``.
    When weights are provided, a weighted fit is performed. Parameter
    descriptions and return values are identical to those in `iter_linear_fit`,
    except returned ``fit`` dictionary does not contain the following
    keys irrelevant to this function: ``'center'``, ``'fitmask'``, and
    ``'eff_nclip'``.

    """
    if xy.size == 0:
        raise NotEnoughPointsError(
            "At least one point is required to find shifts."
        )

    diff_pts = np.subtract(xy, uv, dtype=np.longdouble)

    if wxy is None and wuv is None:
        # no weighting
        w = None

        meanx = diff_pts[:, 0].mean(dtype=np.longdouble)
        meany = diff_pts[:, 1].mean(dtype=np.longdouble)

    else:
        if wxy is None:
            w = np.array(wuv, dtype=np.longdouble)
        elif wuv is None:
            w = np.array(wxy, dtype=np.longdouble)
        else:
            # 1/w = sigma**2 = sigma_xy**2 + sigma_uv**2 = 1/wxy + 1/wuv
            wuv = np.array(wuv, dtype=np.longdouble)
            wxy = np.array(wxy, dtype=np.longdouble)
            m = np.logical_and(wuv > 0, wxy > 0)
            w = np.zeros_like(wuv)
            w[m] = wxy[m] * wuv[m] / (wxy[m] + wuv[m])

        if np.any(w < 0.0):
            raise ValueError("Invalid weights: weights must be non-negative.")

        if not np.sum(w > 0, dtype=int):
            raise ValueError("Not enough valid data for 'shift' fit: "
                             "too many weights are zero!")

        w /= np.sum(w, dtype=np.longdouble)

        meanx = np.dot(w, diff_pts[:, 0])
        meany = np.dot(w, diff_pts[:, 1])

    p = np.array([1.0, 0.0, meanx], dtype=np.longdouble)
    q = np.array([0.0, 1.0, meany], dtype=np.longdouble)

    fit = _build_fit(p, q, 'shift')
    resids = diff_pts - fit['shift']
    fit['resids'] = resids.astype(np.double)
    _compute_stat(fit, residuals=resids, weights=w)
    return fit


# Implementation of geomap 'rscale' fitting based on 'lib/geofit.x'
# by Warren Hack. Support for axis flips and for constraining scale
# added by Mihai Cara.
def fit_rscale(xy, uv, wxy=None, wuv=None, scale=None):
    """ Fits (non-iteratively and without sigma-clipping) a displacement,
    rotation and (optionally) scale transformations between input lists of
    positions ``xy`` and ``uv``. When weights are provided, a weighted fit
    is performed. Parameter descriptions and return values are identical to
    those in `iter_linear_fit`, except returned ``fit`` dictionary does
    not contain the following keys irrelevant to this function: ``'center'``,
    ``'fitmask'``, and ``'eff_nclip'``.

    When ``scale`` is `None` then this function will fit for the scale of the
    transformation. If ``scale`` is a positive number - it will be used as
    fixed scale constraint and it will not be fitted for. A typical usage is
    to set ``scale`` to `None` in order to fit for shifts, rotations *and*
    scale and to set ``scale`` argument to, e.g., 1 to fit for shifts and
    rotations only.

    """
    if len(xy) < 2:
        raise NotEnoughPointsError(
            "At least two points are required to find shifts, rotation, and "
            "scale."
        )

    if scale is None:
        fitgeom = 'rscale'
    elif scale > 0:
        fitgeom = 'rshift'
    else:
        raise ValueError("'scale' argument must be None or a positive number.")

    x = np.array(xy[:, 0], dtype=np.longdouble)
    y = np.array(xy[:, 1], dtype=np.longdouble)
    u = np.array(uv[:, 0], dtype=np.longdouble)
    v = np.array(uv[:, 1], dtype=np.longdouble)

    if wxy is None and wuv is None:
        # no weighting
        w = None

        xm = np.mean(x)
        ym = np.mean(y)
        um = np.mean(u)
        vm = np.mean(v)

        x -= xm
        y -= ym
        u -= um
        v -= vm

        sxv = np.dot(x, v)
        syu = np.dot(y, u)
        sxu = np.dot(x, u)
        syv = np.dot(y, v)
        if scale is None:
            su2 = np.dot(u, u)
            sv2 = np.dot(v, v)
            su2v2 = su2 + sv2

    else:
        if wxy is None:
            w = np.array(wuv, dtype=np.longdouble)
        elif wuv is None:
            w = np.array(wxy, dtype=np.longdouble)
        else:
            # 1/w = sigma**2 = sigma_xy**2 + sigma_uv**2 = 1/wxy + 1/wuv
            wuv = np.array(wuv, dtype=np.longdouble)
            wxy = np.array(wxy, dtype=np.longdouble)
            m = np.logical_and(wuv > 0, wxy > 0)
            w = np.zeros_like(wuv)
            w[m] = wxy[m] * wuv[m] / (wxy[m] + wuv[m])

        if np.any(w < 0.0):
            raise ValueError("Invalid weights: weights must be non-negative.")

        if np.sum(w > 0) < 2:
            raise ValueError(f"Not enough valid data for '{fitgeom:s}' fit: "
                             "too many weights are zero!")

        w /= np.sum(w, dtype=np.longdouble)
        xm = np.dot(w, x)
        ym = np.dot(w, y)
        um = np.dot(w, u)
        vm = np.dot(w, v)

        x -= xm
        y -= ym
        u -= um
        v -= vm

        sxv = np.dot(w, x * v)
        syu = np.dot(w, y * u)
        sxu = np.dot(w, x * u)
        syv = np.dot(w, y * v)
        if scale is None:
            su2 = np.dot(w, u**2)
            sv2 = np.dot(w, v**2)
            su2v2 = su2 + sv2

    det = sxu * syv - sxv * syu
    if det < 0:
        rot_num = sxv + syu
        rot_denom = sxu - syv
    else:
        rot_num = sxv - syu
        rot_denom = sxu + syv

    if rot_num == rot_denom:
        theta = 0.0
    else:
        theta = np.rad2deg(np.arctan2(rot_num, rot_denom))
        if theta < 0:
            theta += 360.0

    ctheta = np.cos(np.deg2rad(theta))
    stheta = np.sin(np.deg2rad(theta))
    s_num = rot_denom * ctheta + rot_num * stheta

    if scale is not None:
        mag = scale
    elif su2v2 > 0.0:
        mag = s_num / su2v2
    else:
        raise SingularMatrixError(
            "Singular matrix: suspected colinear points."
        )

    if det < 0:
        # "flip" y-axis (reflection about x-axis *after* rotation)
        # NOTE: keep in mind that 'matrix' is the transposed rotation matrix.
        sthetax = -mag * stheta
        cthetay = -mag * ctheta
    else:
        sthetax = mag * stheta
        cthetay = mag * ctheta

    cthetax = mag * ctheta
    sthetay = mag * stheta

    sdet = np.sign(det)
    xshift = xm - um * cthetax - sdet * vm * sthetax
    yshift = ym + sdet * um * sthetay - vm * cthetay

    p = np.array([cthetax, sthetay, xshift], dtype=np.longdouble)
    q = np.array([-sthetax, cthetay, yshift], dtype=np.longdouble)

    # Return the shift, rotation, and scale changes
    fit = _build_fit(p, q, fitgeom=fitgeom)
    resids = xy - np.dot(uv, fit['matrix_ld'].T) - fit['shift_ld']
    fit['resids'] = resids.astype(np.double)
    _compute_stat(fit, residuals=resids, weights=w)
    return fit


def fit_rshift(xy, uv, wxy=None, wuv=None):
    """ Fits (non-iteratively and without sigma-clipping) a displacement,
    rotation between input lists of positions ``xy`` and ``uv``. When weights
    are provided, a weighted fit is performed. Parameter descriptions and
    return values are identical to those in `iter_linear_fit`, except
    returned ``fit`` dictionary does not contain the following keys irrelevant
    to this function: ``'center'``, ``'fitmask'``, and ``'eff_nclip'``.

    .. note::
        This function is equivalent to calling `fit_rscale` with ``scale=1``.

    """
    return fit_rscale(xy=xy, uv=uv, wxy=wxy, wuv=wuv, scale=1)


def fit_general(xy, uv, wxy=None, wuv=None):
    """ Fits (non-iteratively and without sigma-clipping) a displacement,
    rotation, scale, and skew transformations (i.e., the full ``2x2``
    transformation matrix) between input lists of positions
    ``xy`` and ``uv``. When weights are provided, a weighted fit is performed.
    Parameter descriptions and return values are identical to those
    in `iter_linear_fit`, except returned ``fit`` dictionary does not contain
    the following keys irrelevant to this function: ``'center'``,
    ``'fitmask'``, and ``'eff_nclip'``.

    """
    if len(xy) < 3:
        raise NotEnoughPointsError(
            "At least three points are required to find 6-parameter linear "
            "affine transformations."
        )

    x = np.array(xy[:, 0], dtype=np.longdouble)
    y = np.array(xy[:, 1], dtype=np.longdouble)
    u = np.array(uv[:, 0], dtype=np.longdouble)
    v = np.array(uv[:, 1], dtype=np.longdouble)

    if wxy is None and wuv is None:
        # no weighting
        w = None

        # Set up products used for computing the fit
        sw = float(x.size)
        sx = x.sum()
        sy = y.sum()
        su = u.sum()
        sv = v.sum()

        sxu = np.dot(x, u)
        syu = np.dot(y, u)
        sxv = np.dot(x, v)
        syv = np.dot(y, v)
        suu = np.dot(u, u)
        svv = np.dot(v, v)
        suv = np.dot(u, v)

    else:
        if wxy is None:
            w = np.array(wuv, dtype=np.longdouble)
        elif wuv is None:
            w = np.array(wxy, dtype=np.longdouble)
        else:
            # 1/w = sigma**2 = sigma_xy**2 + sigma_uv**2 = 1/wxy + 1/wuv
            wuv = np.array(wuv, dtype=np.longdouble)
            wxy = np.array(wxy, dtype=np.longdouble)
            m = np.logical_and(wuv > 0, wxy > 0)
            w = np.zeros_like(wuv)
            w[m] = wxy[m] * wuv[m] / (wxy[m] + wuv[m])

        if np.any(w < 0.0):
            raise ValueError("Invalid weights: weights must be non-negative.")

        if np.sum(w > 0) < 3:
            raise ValueError("Not enough valid data for 'general' fit: "
                             "too many weights are zero!")

        # Set up products used for computing the fit
        sw = np.sum(w, dtype=np.longdouble)
        sx = np.dot(w, x)
        sy = np.dot(w, y)
        su = np.dot(w, u)
        sv = np.dot(w, v)

        sxu = np.dot(w, x * u)
        syu = np.dot(w, y * u)
        sxv = np.dot(w, x * v)
        syv = np.dot(w, y * v)
        suu = np.dot(w, u * u)
        svv = np.dot(w, v * v)
        suv = np.dot(w, u * v)

    m = np.array([[su, sv, sw], [suu, suv, su], [suv, svv, sv]],
                 dtype=np.longdouble)
    a = np.array([sx, sxu, sxv], dtype=np.longdouble)
    b = np.array([sy, syu, syv], dtype=np.longdouble)

    try:
        inv_m = inv(m)
    except np.linalg.LinAlgError:
        raise SingularMatrixError(
            "Singular matrix: suspected colinear points."
        )

    p = np.dot(inv_m, a)
    q = np.dot(inv_m, b)
    if not (np.all(np.isfinite(p)) and np.all(np.isfinite(q))):
        raise SingularMatrixError(
            "Singular matrix: suspected colinear points."
        )  # pragma: no cover

    # Return the shift, rotation, and scale changes
    fit = _build_fit(p, q, 'general')
    resids = xy - np.dot(uv, fit['matrix_ld'].T) - fit['shift_ld']
    fit['resids'] = resids.astype(np.double)
    _compute_stat(fit, residuals=resids, weights=w)
    return fit


def _build_fit(p, q, fitgeom):
    # Build fit matrix:
    fit_matrix = np.vstack((p[:2], q[:2]))

    # determinant of the transformation
    det = p[0] * q[1] - p[1] * q[0]
    sdet = np.sign(det)
    proper = sdet >= 0

    # Create a working copy (no reflections) for computing transformation
    # parameters (scale, rotation angle, skew):
    wfit = fit_matrix.copy()

    # Skew is zero for all fitgeom except 'general':
    skew = 0.0

    if fitgeom == 'shift':
        fit = {
            'shift': np.array([p[2], q[2]], dtype=np.double),
            'shift_ld': np.array([p[2], q[2]], dtype=np.longdouble),
            'matrix': np.array(fit_matrix, dtype=np.double),
            'matrix_ld': np.array(fit_matrix, dtype=np.longdouble),
            'proper_rot': 0.0,
            'rot': (0.0, 0.0),
            '<rot>': 0.0,
            'scale': (1.0, 1.0),
            '<scale>': 1.0,
            'skew': 0.0,
            'proper': proper,
            'fitgeom': 'shift'
        }

        return fit

    # Compute average scale:
    s = np.sqrt(np.abs(det))
    # Compute scales for each axis:
    if fitgeom == 'general':
        sx, sy = np.sqrt(p[:2]**2 + q[:2]**2)
    else:
        sx = s
        sy = s

    # Remove scale from the transformation matrix:
    wfit[:, 0] /= sx
    wfit[:, 1] /= sy

    # Compute rotation angle as if we have a proper rotation.
    # This will also act as *some sort* of "average rotation" even for
    # transformations with different rot_x and rot_y:
    prop_rot = np.rad2deg(
        np.arctan2(wfit[0, 1] - sdet * wfit[1, 0],
                   wfit[0, 0] + sdet * wfit[1, 1])
    )

    if proper and fitgeom in ['rshift', 'rscale']:
        rotx = prop_rot
        roty = prop_rot
        rot = prop_rot

    else:
        rotx = np.rad2deg(np.arctan2(-wfit[1, 0], wfit[0, 0]))
        roty = np.rad2deg(np.arctan2(wfit[0, 1], wfit[1, 1]))
        rot = 0.5 * (rotx + roty)
        skew = np.mod(roty - rotx - 180.0, 360.0) - 180.0

    fit = {
        'shift': np.array([p[2], q[2]], dtype=np.double),
        'shift_ld': np.array([p[2], q[2]], dtype=np.longdouble),
        'matrix': np.array(fit_matrix, dtype=np.double),
        'matrix_ld': np.array(fit_matrix, dtype=np.longdouble),
        'proper_rot': float(prop_rot),
        'rot': (float(rotx), float(roty)),
        '<rot>': float(rot),
        'scale': (float(sx), float(sy)),
        '<scale>': float(s),
        'skew': float(skew),
        'proper': proper,
        'fitgeom': fitgeom
    }

    return fit


def build_fit_matrix(rot, scale=1):
    r"""
    Create an affine transformation matrix (2x2) from the provided rotation
    angle(s) and scale(s):

    .. math::

        M = \begin{bmatrix}
                s_x \cos(\theta_x) & s_y \sin(\theta_y) \\
                -s_x \sin(\theta_x) & s_y \cos(\theta_y)
            \end{bmatrix}

    Parameters
    ----------
    rot: tuple, float, optional
        Rotation angle in degrees. Two values (one for each axis) can be
        provided as a tuple.

    scale: tuple, float, optional
        Scale of the liniar transformation. Two values (one for each axis)
        can be provided as a tuple.

    Returns
    -------
    matrix: numpy.ndarray
       A 2x2 `numpy.ndarray` containing coefficients of a liniear
       transformation.

    """
    if hasattr(rot, '__iter__'):
        rx, ry = map(np.deg2rad, rot)
    else:
        rx = ry = np.deg2rad(float(rot))

    if hasattr(scale, '__iter__'):
        sx, sy = scale
    else:
        sx = sy = float(scale)

    matrix = np.array([[sx * np.cos(rx), sy * np.sin(ry)],
                       [-sx * np.sin(rx), sy * np.cos(ry)]])

    return matrix


class _LevMarLSQFitter2x2(LevMarLSQFitter):
    """ Performs fits of 2D vector-models to 2D reference points. """
    def objective_function(self, fps, *args):
        model, weights, inputs, meas = args
        _fitter_to_model_params(model, fps)
        if weights is None:
            return np.ravel(np.subtract(model(*inputs), meas))
        else:
            return np.ravel(weights * np.subtract(model(*inputs), meas))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A module that provides algorithms matching catalogs and for initial estimation
of shifts based on 2D histograms.

:License: :doc:`../LICENSE`

"""
import logging
from abc import ABC, abstractmethod

import numpy as np
import astropy

from stsci.stimage import xyxymatch

from . import __version__  # noqa: F401

__author__ = 'Mihai Cara'

__all__ = ['MatchCatalogs', 'TPMatch']

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class MatchCatalogs(ABC):
    """ A class that provides common interface for matching catalogs. """

    def __init__(self):
        """
        """

    @abstractmethod
    def __call__(self, refcat, imcat):
        """ Performs catalog matching.

        Parameters
        ----------

        refcat: astropy.table.Table
            A reference source catalog. Reference catalog must contain
            ``'TPx'`` and ``'TPy'`` columns that provide undistorted
            (distortion-correction applied) source coordinates coordinate
            system common (shared) with the image catalog ``imcat``.

        imcat: astropy.table.Table
            Source catalog associated with an image. Image catalog must contain
            ``'TPx'`` and ``'TPy'`` columns that provide undistorted
            (distortion-correction applied) source coordinates coordinate
            system common (shared) with the image catalog ``refcat``.

        Returns
        -------
        (refcat_idx, imcat_idx): tuple of numpy.ndarray
            A tuple of two 1D `numpy.ndarray` containing indices of matched
            sources in the ``refcat`` and ``imcat`` catalogs accordingly.

        """
        pass


class TPMatch(MatchCatalogs):
    """ Catalog source matching in tangent plane. Uses ``xyxymatch``
    algorithm to cross-match sources between this catalog and
    a reference catalog.

    .. note::
        The tangent plane is a plane tangent to the celestial sphere and it
        must be not distorted, that is, if *image* coordinates are distorted,
        then distortion correction must be applied to them before tangent
        plane coordinates are computed. Alternatively, one can think that
        undistorted world coordinates are projected from the sphere onto the
        tangent plane.

    """
    def __init__(self, searchrad=3.0, separation=0.5, use2dhist=True,
                 xoffset=0.0, yoffset=0.0, tolerance=1.0):
        """
        Parameters
        ----------

        searchrad: float, optional
            The search radius for a match (in units of the tangent plane).

        separation: float, optional
            The  minimum  separation in the tangent plane (in units of
            the tangent plane) for sources in the image and reference
            catalogs in order to be considered to be disctinct sources.
            Objects closer together than ``separation`` distance
            are removed from the image and reference coordinate catalogs prior
            to matching. This parameter gets passed directly to
            :py:func:`~stsci.stimage.xyxymatch` for use in matching the object
            lists from each image with the reference catalog's object list.

        use2dhist: bool, optional
            Use 2D histogram to find initial offset?

        xoffset: float, optional
            Initial estimate for the offset in X (in units of the tangent
            plane) between the sources in the image and the reference catalogs.
            This offset will be used for all input images provided.
            This parameter is ignored when ``use2dhist`` is `True`.

        yoffset: float, optional
            Initial estimate for the offset in Y (in units of the tangent
            plane) between the sources in the image and the reference catalogs.
            This offset will be used for all input images provided.
            This parameter is ignored when ``use2dhist`` is `True`.

        tolerance: float, optional
            The matching tolerance (in units of the tangent plane) after
            applying an initial solution derived from the 'triangles'
            algorithm.  This parameter gets passed directly to
            :py:func:`~stsci.stimage.xyxymatch` for use in matching
            the object lists from each image with the reference image's object
            list.

        """
        self._use2dhist = use2dhist

        if searchrad > 0:
            self._searchrad = searchrad
        else:
            raise ValueError("'searchrad' must be a positive number.")

        if separation > 0:
            self._separation = separation
        else:
            raise ValueError("'separation' must be a positive number.")

        if tolerance > 0:
            self._tolerance = tolerance
        else:
            raise ValueError("'tolerance' must be a positive number.")

        self._xoffset = float(xoffset)
        self._yoffset = float(yoffset)

    def __call__(self, refcat, imcat, tp_wcs=None):
        r""" Performs catalog matching.

        Parameters
        ----------

        refcat: astropy.table.Table
            A reference source catalog. When a tangent-plane ``WCS`` is
            provided through ``tp_wcs``, the catalog must contain ``'RA'`` and
            ``'DEC'`` columns which indicate reference source world
            coordinates (in degrees). Alternatively, when ``tp_wcs`` is `None`,
            reference catalog must contain ``'TPx'`` and ``'TPy'`` columns that
            provide undistorted (distortion-correction applied) source
            coordinates in some *tangent plane*. In this case, the ``'RA'``
            and ``'DEC'`` columns in the ``refcat`` catalog will be ignored.

        imcat: astropy.table.Table
            Source catalog associated with an image. Must contain ``'x'`` and
            ``'y'`` columns which indicate source coordinates (in pixels) in
            the associated image. Alternatively, when ``tp_wcs`` is `None`,
            catalog must contain ``'TPx'`` and ``'TPy'`` columns that
            provide undistorted (distortion-correction applied) source
            coordinates in **the same**\ *tangent plane* used to define
            ``refcat``'s tangent plane coordinates. In this case, the ``'x'``
            and ``'y'`` columns in the ``imcat`` catalog will be ignored.

        tp_wcs: TPWCS, None, optional
            A ``WCS`` that defines a tangent plane onto which both
            reference and image catalog sources can be projected. For this
            reason, ``tp_wcs`` is associated with the image in which sources
            from the ``imcat`` catalog were found in the sense that ``tp_wcs``
            must be able to map image coordinates ``'x'`` and ``'y'`` from the
            ``imcat`` catalog to the tangent plane. When ``tp_wcs`` is
            provided, the ``'TPx'`` and ``'TPy'`` columns in both ``imcat`` and
            ``refcat`` catalogs will be ignored (if present).

        Returns
        -------
        (refcat_idx, imcat_idx): tuple of numpy.ndarray
            A tuple of two 1D `numpy.ndarray` containing indices of matched
            sources in the ``refcat`` and ``imcat`` catalogs accordingly.

        """
        # Check catalogs:
        if not isinstance(refcat, astropy.table.Table):
            raise TypeError("'refcat' must be an instance of "
                            "astropy.table.Table")

        if not refcat:
            raise ValueError("Reference catalog must contain at least one "
                             "source.")

        if not isinstance(imcat, astropy.table.Table):
            raise TypeError("'imcat' must be an instance of "
                            "astropy.table.Table")

        if not imcat:
            raise ValueError("Image catalog must contain at least one "
                             "source.")

        if tp_wcs is None:
            if 'TPx' not in refcat.colnames or 'TPy' not in refcat.colnames:
                raise KeyError("When tangent plane WCS is not provided, "
                               "'refcat' must contain both 'TPx' and 'TPy' "
                               "columns.")

            if 'TPx' not in imcat.colnames or 'TPy' not in imcat.colnames:
                raise KeyError("When tangent plane WCS is not provided, "
                               "'imcat' must contain both 'TPx' and 'TPy' "
                               "columns.")

            imxy = np.asarray([imcat['TPx'], imcat['TPy']]).T
            refxy = np.asarray([refcat['TPx'], refcat['TPy']]).T

        else:
            if 'RA' not in refcat.colnames or 'DEC' not in refcat.colnames:
                raise KeyError("When tangent plane WCS is provided,  'refcat' "
                               "must contain both 'RA' and 'DEC' columns.")

            if 'x' not in imcat.colnames or 'y' not in imcat.colnames:
                raise KeyError("When tangent plane WCS is provided,  'imcat' "
                               "must contain both 'x' and 'y' columns.")

            # compute x & y in the tangent plane provided by tp_wcs:
            imxy = np.asarray(
                tp_wcs.det_to_tanp(imcat['x'], imcat['y'])
            ).T

            refxy = np.asarray(
                tp_wcs.world_to_tanp(refcat['RA'], refcat['DEC'])
            ).T

        imcat_name = imcat.meta.get('name', 'Unnamed')
        if imcat_name is None:
            imcat_name = 'Unnamed'

        refcat_name = refcat.meta.get('name', 'Unnamed')
        if refcat_name is None:
            refcat_name = 'Unnamed'

        log.info("Matching sources from '{:s}' catalog with sources from the "
                 "reference '{:s}' catalog."
                 .format(imcat_name, refcat_name))

        ps = 1.0 if tp_wcs is None else tp_wcs.tanp_center_pixel_scale

        if self._use2dhist:
            # Determine xyoff (X,Y offset) and tolerance
            # to be used with xyxymatch:
            zpxoff, zpyoff = _estimate_2dhist_shift(
                imxy / ps,
                refxy / ps,
                searchrad=self._searchrad
            )
            xyoff = (zpxoff * ps, zpyoff * ps)

        else:
            xyoff = (self._xoffset * ps, self._yoffset * ps)

        matches = xyxymatch(
            imxy,
            refxy,
            origin=xyoff,
            tolerance=ps * self._tolerance,
            separation=ps * self._separation
        )

        return matches['ref_idx'], matches['input_idx']


def _xy_2dhist(imgxy, refxy, r):
    # This code replaces the C version (arrxyzero) from carrutils.c
    # It is about 5-8 times slower than the C version.
    dx = np.subtract.outer(imgxy[:, 0], refxy[:, 0]).ravel()
    dy = np.subtract.outer(imgxy[:, 1], refxy[:, 1]).ravel()
    idx = np.where((dx < r + 0.5) & (dx >= -r - 0.5) &
                   (dy < r + 0.5) & (dy >= -r - 0.5))
    r = int(np.ceil(r))
    h = np.histogram2d(dx[idx], dy[idx], 2 * r + 1,
                       [[-r - 0.5, r + 0.5], [-r - 0.5, r + 0.5]])
    return h[0].T


def _estimate_2dhist_shift(imgxy, refxy, searchrad=3.0):
    """ Create a 2D matrix-histogram which contains the delta between each
        XY position and each UV position. Then estimate initial offset
        between catalogs.
    """
    log.info("Computing initial guess for X and Y shifts...")

    # create ZP matrix
    zpmat = _xy_2dhist(imgxy, refxy, r=searchrad)

    nonzeros = np.count_nonzero(zpmat)
    if nonzeros == 0:
        # no matches within search radius. Return (0, 0):
        log.warning("No matches found within a search radius of {:g} pixels."
                    .format(searchrad))
        return 0.0, 0.0

    elif nonzeros == 1:
        # only one non-zero bin:
        yp, xp = np.unravel_index(np.argmax(zpmat), zpmat.shape)
        maxval = zpmat[yp, xp]
        xp -= searchrad
        yp -= searchrad
        log.info("Found initial X and Y shifts of {:.4g}, {:.4g} "
                 "based on a single non-zero bin and {} matches"
                 .format(xp, yp, int(maxval)))
        return xp, yp

    (xp, yp), fit_status, fit_sl = _find_peak(zpmat, peak_fit_box=5,
                                              mask=zpmat > 0)
    if fit_status.startswith('ERROR'):
        log.warning("No valid shift found within a search radius of {:g} "
                    "pixels.".format(searchrad))
        return 0.0, 0.0

    xp -= searchrad
    yp -= searchrad

    if fit_status == 'WARNING:EDGE':
        log.info("Found peak in the 2D histogram lies at the edge of the "
                 "histogram. Try increasing 'searchrad' for improved results.")

    flux = int(zpmat[fit_sl].sum())

    # Attempt to estimate "significance of detection":
    maxval = zpmat.max()
    zpmat_mask = (zpmat > 0) & (zpmat < maxval)

    bkg = zpmat[zpmat_mask].mean() if np.any(zpmat_mask) else -1.0

    if bkg > 0:  # pragma: no branch
        bkg = zpmat[zpmat_mask].mean()
        sig = maxval / np.sqrt(bkg)
        log.info("Found initial X and Y shifts of {:.4g}, {:.4g} "
                 "with significance of {:.4g} and {:d} matches."
                 .format(xp, yp, sig, flux))

    else:
        log.warning("Unable to estimate significance of the detection of the "
                    "initial shift.")
        log.info("Found initial X and Y shifts of {:.4g}, {:.4g} "
                 "with {:d} matches."
                 .format(xp, yp, flux))

    return xp, yp


def _find_peak(data, peak_fit_box=5, mask=None):
    """
    Find location of the peak in an array. This is done by fitting a second
    degree 2D polynomial to the data within a `peak_fit_box` and computing the
    location of its maximum. An initial estimate of the position of the maximum
    will be performed by searching for the location of the pixel/array element
    with the maximum value.

    Parameters
    ----------
    data: numpy.ndarray
        2D data.

    peak_fit_box: int, optional
        Size (in pixels) of the box around the initial estimate of the maximum
        to be used both for quadratic fitting from which peak location is
        computed and for the center-of-mass estimate.
        It is assumed that fitting box is a square with sides of length
        given by ``peak_fit_box``.

    mask: numpy.ndarray, optional
        A boolean type `~numpy.ndarray` indicating "good" pixels in image data
        (`True`) and "bad" pixels (`False`). If not provided all pixels
        in `image_data` will be used for fitting.

    Returns
    -------
    coord: tuple of float
        A pair of coordinates of the peak.

    fit_status: str
        Status of the peak search. Currently the following values can be
        returned:

        - ``'SUCCESS'``: Fit was successful and peak is not on the edge of
          the input array;
        - ``'ERROR:NODATA'``: Not enough valid data to perform the fit; The
          returned coordinate is the center of input array;
        - ``'WARNING:EDGE'``: Peak lies on the edge of the input array.
          Returned coordinates are the result of a discreet search;
        - ``'WARNING:BADFIT'``: Performed fid did not find a maximum or the
          estimated maximum is outside of the fit box. Returned
          coordinates are the result of either a center-of-mass estimate or
          a discreet search;
        - ``'WARNING:CENTER-OF-MASS'``: Returned coordinates are the result
          of a center-of-mass estimate instead of a polynomial fit. This is
          either due to too few points to perform a fit, polynomial peak being
          outside of the fit box, or due to a failure of the polynomial fit.

    fit_box: a tuple of `slice`
        A tuple of `slice` objects of the form
        ``(slice(y1, y2, None), slice(x1, x2, None))`` that
        indicates pixel ranges used for fitting (these indices can be used
        directly for slicing input data)

    """

    def _center_of_mass(v, d, x1, x2, y1, y2):
        # Compute center-of-mass. Assumes that ``v`` was computed using
        # coordinates ``x``, ``y`` relative to ``x1 - 1`` and ``y1 - 1``.
        # Returned coordinate is relative to origin.
        m = np.logical_not(np.isfinite(d))
        vx = v[:, 1].flatten()
        vy = v[:, 2].flatten()
        d[m] = 0
        vx[m] = 0
        vy[m] = 0
        dt = d.sum()
        if dt == 0.0:
            coord = ((x2 + x1 - 1.0) / 2.0, (y2 + y1 - 1.0) / 2.0)
            return coord, 'ERROR:NODATA'
        xc = np.dot(vx, d) / dt
        yc = np.dot(vy, d) / dt
        return ((float(x1 + xc - 1), float(y1 + yc - 1)),
                'WARNING:CENTER-OF-MASS')

    # check arguments:
    if peak_fit_box < 1:
        raise ValueError("peak_fit_box must be at least 1 pixel in size.")
    data = np.asarray(data, dtype=np.double)
    ny, nx = data.shape

    # find index of the pixel having maximum value:
    finite = np.isfinite(data)
    mask = finite if mask is None else np.logical_and(mask, finite)

    j, i = np.indices(data.shape)
    i = i[mask]
    j = j[mask]

    if i.size == 0:
        # no valid data:
        coord = ((nx - 1.0) / 2.0, (ny - 1.0) / 2.0)
        return coord, 'ERROR:NODATA', np.s_[0:ny, 0:nx]

    ind = np.argmax(data[mask])
    imax = i[ind]
    jmax = j[ind]
    coord = (float(imax), float(jmax))

    if data[jmax, imax] < 1:
        # no valid data: we need some counts in the histogram bins
        coord = ((nx - 1.0) / 2.0, (ny - 1.0) / 2.0)
        return coord, 'ERROR:NODATA', np.s_[0:ny, 0:nx]

    # choose a box around maxval pixel:
    x1 = max(0, imax - peak_fit_box // 2)
    x2 = min(nx, x1 + peak_fit_box)
    y1 = max(0, jmax - peak_fit_box // 2)
    y2 = min(ny, y1 + peak_fit_box)

    # if peak is at the edge of the box, return integer indices of the max:
    if imax == x1 or imax == x2 - 1 or jmax == y1 or jmax == y2 - 1:
        return coord, 'WARNING:EDGE', np.s_[y1:y2, x1:x2]

    # expand the box if needed:
    if (x2 - x1) < peak_fit_box:  # pragma: no branch
        if x1 == 0:  # pragma: no branch
            x2 = min(nx, x1 + peak_fit_box)
        if x2 == nx:  # pragma: no branch
            x1 = max(0, x2 - peak_fit_box)

    if (y2 - y1) < peak_fit_box:  # pragma: no branch
        if y1 == 0:  # pragma: no branch
            y2 = min(ny, y1 + peak_fit_box)
        if y2 == ny:  # pragma: no branch
            y1 = max(0, y2 - peak_fit_box)

    assert x2 - x1 > 0 or y2 - y1 > 0

    fit_slice = np.s_[y1:y2, x1:x2]

    # fit a 2D 2nd degree polynomial to data:
    m = mask[fit_slice].ravel()
    xi = np.arange(x1, x2) - (x1 - 1)
    yi = np.arange(y1, y2) - (y1 - 1)
    x, y = np.meshgrid(xi, yi)
    x = x.ravel()
    y = y.ravel()
    v = np.vstack((np.ones_like(x), x, y, x * y, x * x, y * y)).T[m]
    d = data[fit_slice].ravel()[m]

    if d.size < 6:
        # we need at least 6 points to fit a 2D quadratic polynomial
        # attempt center-of-mass instead:
        coord, fit_status = _center_of_mass(v, d, x1, x2, y1, y2)
        return coord, fit_status, fit_slice

    try:
        c = np.linalg.lstsq(v, d, rcond=None)[0]
        if not np.all(np.isfinite(c)):
            raise np.linalg.LinAlgError("Results of the fit are not finite.")
    except np.linalg.LinAlgError as e:
        log.warning("Least squares failed!\n{}".format(e))

        # attempt center-of-mass instead:
        coord, fit_status = _center_of_mass(v, d, x1, x2, y1, y2)
        return coord, fit_status, fit_slice

    # find maximum of the polynomial:
    _, c10, c01, c11, c20, c02 = c
    det = 4 * c02 * c20 - c11**2
    if det <= 0 or ((c20 > 0.0 and c02 >= 0.0) or (c20 >= 0.0 and c02 > 0.0)):
        # polynomial does not have max. return maximum value in the data:
        coord, fit_status = _center_of_mass(v, d, x1, x2, y1, y2)
        if fit_status.startswith('ERROR'):
            return coord, fit_status, fit_slice
        return coord, 'WARNING:BADFIT', fit_slice

    xm = (c01 * c11 - 2.0 * c02 * c10) / det + x1 - 1
    ym = (c10 * c11 - 2.0 * c01 * c20) / det + y1 - 1

    if x1 <= xm <= (x2 - 1.0) and y1 <= ym <= (y2 - 1.0):
        coord = (xm, ym)
        fit_status = 'SUCCESS'

    else:
        coord, fit_status = _center_of_mass(v, d, x1, x2, y1, y2)

    return coord, fit_status, fit_slice



import pytest

from astropy import wcs as fitswcs
from astropy.table import Table

from tweakwcs.linearfit import build_fit_matrix
from . helper_tpwcs import make_mock_jwst_wcs


@pytest.fixture(scope='module')
def mock_jwst_wcs():
    cd = build_fit_matrix((36, 47), 1e-5)
    w = make_mock_jwst_wcs(
        v2ref=123.0, v3ref=500.0, roll=115.0, crpix=[512.0, 512.0],
        cd=cd, crval=[82.0, 12.0]
    )
    return w


@pytest.fixture(scope='function')
def mock_fits_wcs():
    cd = build_fit_matrix((36, 47), 1e-5)
    w = fitswcs.WCS(naxis=2)
    w.wcs.cd = cd
    w.wcs.crval = [82.0, 12.0]
    w.wcs.crpix = [512.0, 512.0]
    w.wcs.ctype = ['RA---TAN', 'DEC--TAN']
    w.pixel_shape = [1024, 2048]
    w.pixel_bounds = ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5))
    w.wcs.set()
    return w


@pytest.fixture(scope='function')
def empty_imcat():
    imcat = Table([[], []], names=['x', 'y'])
    return imcat


@pytest.fixture(scope='function')
def empty_refcat():
    refcat = Table([[], []], names=['RA', 'DEC'])
    return refcat



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides support for working with image footprints on the sky and
source catalogs.

:Authors: Mihai Cara

:License: :doc:`../LICENSE`

"""
import os
import logging
import numbers
from copy import deepcopy
from distutils.version import LooseVersion

import numpy as np
from astropy import table
from astropy.utils.decorators import deprecated_renamed_argument
from spherical_geometry.polygon import SphericalPolygon, MalformedPolygonError

from . linearfit import SUPPORTED_FITGEOM_MODES

try:
    import gwcs
    if LooseVersion(gwcs.__version__) > '0.12.0':
        from gwcs.geometry import CartesianToSpherical, SphericalToCartesian
        _GWCS_VER_GT_0P12 = True
    else:
        _GWCS_VER_GT_0P12 = False

except ImportError:
    _GWCS_VER_GT_0P12 = False


if _GWCS_VER_GT_0P12:
    _S2C = SphericalToCartesian(name='s2c', wrap_lon_at=180)
    _C2S = CartesianToSpherical(name='c2s', wrap_lon_at=180)

else:
    def _S2C(phi, theta):
        phi = np.deg2rad(phi)
        theta = np.deg2rad(theta)
        cs = np.cos(theta)
        x = cs * np.cos(phi)
        y = cs * np.sin(phi)
        z = np.sin(theta)
        return x, y, z

    def _C2S(x, y, z):
        h = np.hypot(x, y)
        phi = np.rad2deg(np.arctan2(y, x))
        theta = np.rad2deg(np.arctan2(z, h))
        return phi, theta

from .wcsutils import planar_rot_3d
from .tpwcs import TPWCS
from .linalg import inv
from .linearfit import iter_linear_fit

from . import __version__  # noqa: F401

__author__ = 'Mihai Cara'

__all__ = ['convex_hull', 'RefCatalog', 'WCSImageCatalog', 'WCSGroupCatalog']


# _S2C = SphericalToCartesian(name='s2c', wrap_lon_at=180)
# _C2S = CartesianToSpherical(name='c2s', wrap_lon_at=180)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def _is_int(n):
    return (
        (isinstance(n, numbers.Integral) and not isinstance(n, bool)) or
        (isinstance(n, np.generic) and np.issubdtype(n, np.integer))
    )


class WCSImageCatalog(object):
    """
    A class that holds information pertinent to an image WCS and a source
    catalog of the sources found in that image.

    .. warning::
        If ``tpwcs.meta`` dictionary contains any of the following
        keywords ``'catalog'``, ``'name'``, or ``'group_id'``, they
        will be ignored without warning.

    """

    def __init__(self, catalog, tpwcs, name=None, group_id=None, meta={}):
        """
        Parameters
        ----------
        catalog: astropy.table.Table
            Source catalog associated with an image. Must contain ``'x'`` and
            ``'y'`` columns which indicate source coordinates (in pixels) in
            the associated image.

        tpwcs: TPWCS
            ``TPWCS``-derived tangent-plane WCS corrector object associated
            with the image from which the catalog was derived.

        name: str, None, optional
            Image catalog's name. This is used to identify catalog during
            logging. If ``name`` is `None`, the ``name`` of this
            ``WCSImageCatalog`` object will be set to ``'Unknown'``.

        group_id: hashable, None, optional
            Group ID that may be used for identifying catalogs that need
            to be aligned together. ``group_id`` must be hashable.

        meta: dict, optional
            Additional information about image, catalog, and/or WCS to be
            stored (for convenience) within `WCSImageCatalog` object.

        """
        self.name = name
        self.group_id = group_id

        self._catalog = None
        self._bb_radec = None

        self.img_bounding_ra = None
        self.img_bounding_dec = None

        self.meta = dict(meta)
        self._fit_info = {'status': 'SKIPPED'}

        self.tpwcs = tpwcs
        self.catalog = catalog

    @property
    def tpwcs(self):
        """ Get :py:class:`TPWCS` WCS. """
        return self._tpwcs

    @tpwcs.setter
    def tpwcs(self, tpwcs):
        """ Get/Set catalog's WCS (a :py:class:`TPWCS`-derived object).

        .. note::
            Setting the WCS triggers automatic bounding polygon recalculation.

        Parameters
        ----------
        tpwcs: TPWCS
            ``TPWCS``-derived tangent-plane WCS corrector object associated
            with the image from which the catalog was extracted.

        """
        if not isinstance(tpwcs, TPWCS):
            raise TypeError("Unsupported 'tpwcs' type. "
                            "'tpwcs' must be a subtype of TPWCS.")
        self._tpwcs = tpwcs

        # create spherical polygon bounding the image
        self.calc_bounding_polygon()

    @property
    def name(self):
        """
        Get/set catalog's name. This is used to identify catalog during
        logging. Upon setting, the value will be converted to a `str`.
        When setting to `None`, the ``name`` will be set to ``'Unknown'``.

        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = 'Unknown' if name is None else str(name)

    @property
    def group_id(self):
        """
        Get/set :py:class:`WCSImageCatalog` object's group ID that may be used
        for identifying catalogs that need to be aligned together.
        ``group_id`` must be hashable.

        """
        return self._group_id

    @group_id.setter
    def group_id(self, group_id):
        # check if input is hashable:
        {group_id: None}
        self._group_id = group_id

    @property
    def fit_status(self):
        """
        Get/Set fit status. This property is a shortcut to the ``'status'``
        key value in the ``fit_info`` dictionary. When the
        :py:class:`WCSImageCatalog` object is created, ``fit_status`` is
        initially set to ``'SKIPPED'``. Alignment tools are reponsible for
        updating catalog's fit status.

        """
        return self._fit_info['status']

    @fit_status.setter
    def fit_status(self, fit_status):
        self._fit_info['status'] = fit_status

    @property
    def fit_info(self):
        """
        Get fit information - a dictionary. This class sets only the
        ``'status'`` field but fitting routines may set additional fields.

        """
        return self._fit_info

    @property
    def catalog(self):
        """ Get/set image's catalog. """
        return self._catalog

    @catalog.setter
    def catalog(self, catalog):
        if catalog is None:
            self._catalog = None
            return

        if 'x' not in catalog.colnames or 'y' not in catalog.colnames:
            raise ValueError("An image catalog must contain 'x' and 'y' "
                             "columns!")

        self._catalog = table.Table(catalog.copy(), masked=True)
        self._catalog.meta['name'] = self._name

        if 'id' not in self._catalog.colnames:  # pragma: no branch
            self._catalog['id'] = np.arange(1, len(self._catalog) + 1)

        # create spherical polygon bounding the image
        self.calc_bounding_polygon()

    def det_to_world(self, x, y):
        """
        Convert pixel coordinates to sky coordinates using full
        (i.e., including distortions) transformations.

        """
        return self._tpwcs.det_to_world(x, y)

    def world_to_det(self, ra, dec):
        """
        Convert sky coordinates to image's pixel coordinates using full
        (i.e., including distortions) transformations.

        """
        return self._tpwcs.world_to_det(ra, dec)

    def det_to_tanp(self, x, y):
        """
        Convert detector (pixel) coordinates to tangent plane coordinates.

        """
        return self._tpwcs.det_to_tanp(x, y)

    def tanp_to_det(self, x, y):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        return self._tpwcs.tanp_to_det(x, y)

    def tanp_to_world(self, x, y):
        """
        Convert tangent plane coordinates to world coordinates.

        """
        return self._tpwcs.tanp_to_world(x, y)

    def world_to_tanp(self, ra, dec):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        return self._tpwcs.world_to_tanp(ra, dec)

    @property
    def polygon(self):
        """ Get image's (or catalog's) bounding spherical polygon.
        """
        return self._polygon

    def intersection(self, wcsim):
        """
        Compute intersection of this `WCSImageCatalog` object and another
        `WCSImageCatalog`, `WCSGroupCatalog`, or
        :py:class:`~spherical_geometry.polygon.SphericalPolygon`
        object.

        Parameters
        ----------
        wcsim: WCSImageCatalog, WCSGroupCatalog, SphericalPolygon
            Another object that should be intersected with this
            `WCSImageCatalog`.

        Returns
        -------
        polygon: SphericalPolygon
            A :py:class:`~spherical_geometry.polygon.SphericalPolygon` that is
            the intersection of this `WCSImageCatalog` and `wcsim`.

        """
        if isinstance(wcsim, (WCSImageCatalog, WCSGroupCatalog)):
            return self._polygon.intersection(wcsim.polygon)
        else:
            return self._polygon.intersection(wcsim)

    # TODO: due to a bug in the sphere package, see
    #       https://github.com/spacetelescope/sphere/issues/74
    #       intersections with polygons formed as union does not work.
    #       For this reason I re-implement 'intersection_area' below with
    #       a workaround for the bug.
    #       The original implementation should be uncommented once the bug
    #       is fixed.
    #
    # def intersection_area(self, wcsim):
        # """ Calculate the area of the intersection polygon.
        # """
        # return np.fabs(self.intersection(wcsim).area())
    def intersection_area(self, wcsim):
        """ Calculate the area of the intersection polygon. """
        if isinstance(wcsim, (WCSImageCatalog, RefCatalog)):
            return np.fabs(self.intersection(wcsim).area())

        else:
            # this is bug workaround for image groups (multi-unions):
            area = 0.0
            for wim in wcsim:
                area += np.fabs(
                    self.polygon.intersection(wim.polygon).area()
                )
            return area

    def _guarded_intersection_area(self, wcsim):
        """
        Calculate the area of the intersection polygon. If some
        intersections fail due to a bug/limitation of ``spherical_geometry``
        then the area of the valid intersections will be returned.
        If images do not intersect or intersection fails, 0 will be returned.
        """
        if isinstance(wcsim, (WCSImageCatalog, RefCatalog)):
            try:
                return np.fabs(self.intersection(wcsim).area())
            except MalformedPolygonError:
                return 0.0

        else:
            # this is bug workaround:
            area = 0.0
            for wim in wcsim:
                try:
                    area += np.fabs(
                        self.polygon.intersection(wim.polygon).area()
                    )
                except MalformedPolygonError:
                    continue
            return area

    def _calc_chip_bounding_polygon(self, stepsize=None):
        """
        Compute image's bounding polygon.

        Parameters
        ----------
        stepsize: int, None, optional
            Indicates the maximum separation between two adjacent vertices
            of the bounding polygon along each side of the image. Corners
            of the image are included automatically. If `stepsize` is `None`,
            bounding polygon will contain only vertices of the image.

        """
        if self.tpwcs is None or self.catalog is None:
            return

        if self.tpwcs.bounding_box is None:
            # just take max image coordinates from catalogs as bounds:
            lx = -0.5
            ly = -0.5
            hx = max(1, int(np.ceil(np.amax(self._catalog['x'])))) - 0.5
            hy = max(1, int(np.ceil(np.amax(self._catalog['y'])))) - 0.5

        else:
            ((lx, hx), (ly, hy)) = self.tpwcs.bounding_box

        if stepsize is None:
            nintx = 2
            ninty = 2
        else:
            nintx = max(2, int(np.ceil((hx - lx) / stepsize)))
            ninty = max(2, int(np.ceil((hy - ly) / stepsize)))

        xs = np.linspace(lx, hx, nintx, dtype=np.double)
        ys = np.linspace(ly, hy, ninty, dtype=np.double)[1:-1]
        nptx = xs.size
        npty = ys.size

        npts = 2 * (nptx + npty)

        borderx = np.empty((npts + 1,), dtype=np.double)
        bordery = np.empty((npts + 1,), dtype=np.double)

        # "bottom" points:
        borderx[:nptx] = xs
        bordery[:nptx] = ly
        # "right"
        sl = np.s_[nptx:nptx + npty]
        borderx[sl] = hx
        bordery[sl] = ys
        # "top"
        sl = np.s_[nptx + npty:2 * nptx + npty]
        borderx[sl] = xs[::-1]
        bordery[sl] = hy
        # "left"
        sl = np.s_[2 * nptx + npty:-1]
        borderx[sl] = lx
        bordery[sl] = ys[::-1]

        # close polygon:
        borderx[-1] = borderx[0]
        bordery[-1] = bordery[0]

        ra, dec = self.det_to_world(borderx, bordery)
        # TODO: for strange reasons, occasionally ra[0] != ra[-1] and/or
        #       dec[0] != dec[-1] (even though we close the polygon in the
        #       previous two lines). Then SphericalPolygon fails because
        #       points are not closed. Therefore we force it to be closed:
        ra[-1] = ra[0]
        dec[-1] = dec[0]

        self.img_bounding_ra = ra
        self.img_bounding_dec = dec
        self._bb_radec = (ra, dec)
        self._polygon = SphericalPolygon.from_radec(ra, dec)

    def _calc_cat_convex_hull(self):
        """
        Compute convex hull that bounds the sources in the catalog.

        """
        if self.tpwcs is None or self.catalog is None:
            return

        x = self.catalog['x']
        y = self.catalog['y']

        if len(x) == 0:
            # no points
            raise RuntimeError(  # pragma: no cover
                "Unexpected error: Contact software developer"
            )

        elif len(x) > 2:
            ra, dec = convex_hull(x, y, wcs=self.det_to_world)
            # else, for len(x) in [1, 2], use entire image footprint.
            # TODO: a more robust algorithm should be implemented to deal with
            #       len(x) in [1, 2] cases.

            # TODO: for strange reasons, occasionally ra[0] != ra[-1] and/or
            #       dec[0] != dec[-1] (even though we close the polygon in the
            #       previous two lines). Then SphericalPolygon fails because
            #       points are not closed. Threfore we force it to be closed:
            ra[-1] = ra[0]
            dec[-1] = dec[0]

            self._bb_radec = (ra, dec)
            self._polygon = SphericalPolygon.from_radec(ra, dec)
            self._poly_area = np.fabs(self._polygon.area())

    def calc_bounding_polygon(self):
        """
        Calculate bounding polygon of the image or of the sources in the
        catalog (if catalog was set).

        """
        # we need image's footprint for later:
        self._calc_chip_bounding_polygon()

        # create smallest convex spherical polygon bounding all sources:
        if self.catalog:
            self._calc_cat_convex_hull()

    @property
    def bb_radec(self):
        """
        Get a 2xN `numpy.ndarray` of RA and DEC of the vertices of the
        bounding polygon.

        """
        return self._bb_radec


class WCSGroupCatalog(object):
    """
    A class that holds together `WCSImageCatalog` image catalog objects
    whose relative positions are fixed and whose source catalogs should be
    fitted together to a reference catalog.

    """

    def __init__(self, images, name=None):
        """
        Parameters
        ----------
        images: list of WCSImageCatalog
            A list of `WCSImageCatalog` image catalogs.

        name: str, None, optional
            Name of the group.

        """
        self._catalog = None

        if isinstance(images, WCSImageCatalog):
            self._images = [images]
            if images.catalog is None:
                raise ValueError("Each input WCS image catalog must have a "
                                 "valid catalog.")

        elif hasattr(images, '__iter__'):
            if not images:
                raise ValueError("List of images cannot be empty.")

            self._images = []
            for im in images:
                if not isinstance(im, WCSImageCatalog):
                    raise TypeError("Each element of the 'images' parameter "
                                    "must be an 'WCSImageCatalog' object.")
                if im.catalog is None:
                    raise ValueError("Each input WCS image catalog must have "
                                     "a valid catalog.")
                self._images.append(im)

        else:
            raise TypeError("Parameter 'images' must be either a single "
                            "'WCSImageCatalog' object or a list of "
                            "'WCSImageCatalog' objects")

        self._name = name
        self._catalog = self.create_group_catalog()
        self.update_bounding_polygon()

    @property
    def name(self):
        """ Get/set :py:class:`WCSImageCatalog` object's name.
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def polygon(self):
        """ Get image's (or catalog's) bounding spherical polygon.
        """
        return self._polygon

    def intersection(self, wcsim):
        """
        Compute intersection of this `WCSGroupCatalog` object and another
        `WCSImageCatalog`, `WCSGroupCatalog`, or
        :py:class:`~spherical_geometry.polygon.SphericalPolygon`
        object.

        Parameters
        ----------
        wcsim: WCSImageCatalog, WCSGroupCatalog, SphericalPolygon
            Another object that should be intersected with this
            `WCSGroupCatalog`.

        Returns
        -------
        polygon: SphericalPolygon
            A :py:class:`~spherical_geometry.polygon.SphericalPolygon` that is
            the intersection of this `WCSGroupCatalog` and `wcsim`.

        """
        if isinstance(wcsim, (WCSGroupCatalog, WCSImageCatalog)):
            return self._polygon.intersection(wcsim.polygon)
        else:
            return self._polygon.intersection(wcsim)

    # TODO: due to a bug in the sphere package, see
    #       https://github.com/spacetelescope/sphere/issues/74
    #       intersections with polygons formed as union does not work.
    #       For this reason I re-implement 'intersection_area' below with
    #       a workaround for the bug.
    #       The original implementation should be uncommented once the bug
    #       is fixed.
    #
    # def intersection_area(self, wcsim):
        # """ Calculate the area of the intersection polygon.
        # """
        # return np.fabs(self.intersection(wcsim).area())
    def intersection_area(self, wcsim):
        """ Calculate the area of the intersection polygon.
        """
        return sum(im.intersection_area(wcsim) for im in self._images)

    def _guarded_intersection_area(self, wcsim):
        """
        Calculate the area of the intersection polygon. If some
        intersections fail due to a bug/limitation of ``spherical_geometry``
        then the area of the valid intersections will be returned.
        If images do not intersect or intersection fails, 0 will be returned.
        """
        return sum(im._guarded_intersection_area(wcsim) for im in self._images)

    def update_bounding_polygon(self):
        """ Recompute bounding polygons of the member images.
        """
        polygons = [im.polygon for im in self._images]
        if polygons:
            try:
                self._polygon = SphericalPolygon.multi_union(polygons)
            except MalformedPolygonError:
                log.warning(
                    "MalformedPolygonError in spherical_geometry. Using "
                    "convex hull instead of multi_union. Alignment order "
                    "may be sub-optimal."
                )
                refcat = RefCatalog(self._catalog)
                self._polygon = refcat.polygon

        else:
            self._polygon = SphericalPolygon([])

    def __len__(self):
        return len(self._images)

    def __getitem__(self, idx):
        return self._images[idx]

    def __iter__(self):
        for image in self._images:
            yield image

    @property
    def catalog(self):
        """ Get/set image's catalog.
        """
        return self._catalog

    def create_group_catalog(self):
        """
        Combine member's image catalogs into a single group's catalog.

        Returns
        -------
        group_catalog: astropy.table.Table
            Combined group catalog.

        """
        catalogs = []
        catno = 0
        has_weights = None
        cat_names = []
        for image in self._images:
            catlen = len(image.catalog)
            if catlen == 0:
                continue

            cat_name = image.catalog.meta.get(
                'name',
                image.name if image.name else 'Unnamed'
            )
            cat_names.append(cat_name)

            if has_weights is None:
                has_weights = 'weight' in image.catalog.colnames
            elif has_weights != ('weight' in image.catalog.colnames):
                raise KeyError("Non-empty catalogs in a group must all "
                               "either have or not have 'weight' column.")

            if image.name is None:
                catname = 'Catalog #{:d}'.format(catno)
            else:
                catname = image.name

            col_catname = table.MaskedColumn(catlen * [catname],
                                             name='cat_name')
            col_imcatidx = table.MaskedColumn(catlen * [catno],
                                              name='_imcat_idx')
            col_id = table.MaskedColumn(image.catalog['id'])
            col_x = table.MaskedColumn(image.catalog['x'], dtype=np.double)
            col_y = table.MaskedColumn(image.catalog['y'], dtype=np.double)
            ra, dec = image.det_to_world(
                image.catalog['x'], image.catalog['y']
            )
            col_ra = table.MaskedColumn(ra, dtype=np.double, name='RA')
            col_dec = table.MaskedColumn(dec, dtype=np.double, name='DEC')

            if has_weights:
                col_wght = table.MaskedColumn(image.catalog['weight'],
                                              dtype=np.double)

                cat = table.Table(
                    [col_imcatidx, col_catname, col_id, col_x,
                     col_y, col_ra, col_dec, col_wght],
                    masked=True
                )

            else:
                cat = table.Table(
                    [col_imcatidx, col_catname, col_id, col_x,
                     col_y, col_ra, col_dec],
                    masked=True
                )

            catalogs.append(cat)
            catno += 1

        catname = os.path.commonprefix(cat_names) if cat_names else None

        if catno:
            cat = table.vstack(catalogs, join_type='exact')

        else:
            # no catalogs with sources. Create an empty table with required
            # columns and types:
            image = self._images[0]
            if image.name is None:
                catname = 'Catalog #{:d}'.format(catno)
            else:
                catname = image.name

            col_catname = table.MaskedColumn([catname], name='cat_name')
            col_catname = col_catname[[False]]
            col_imcatidx = table.MaskedColumn([], dtype=int,
                                              name='_imcat_idx')
            col_id = table.MaskedColumn(image.catalog['id'])
            col_x = table.MaskedColumn([], name='x', dtype=np.double)
            col_y = table.MaskedColumn([], name='y', dtype=np.double)
            col_ra = table.MaskedColumn([], name='RA', dtype=np.double)
            col_dec = table.MaskedColumn([], name='DEC', dtype=np.double)

            cat = table.Table(
                [col_imcatidx, col_catname, col_id, col_x,
                 col_y, col_ra, col_dec],
                masked=True
            )

        if catname:
            cat.meta['name'] = catname

        return cat

    def get_unmatched_cat(self):
        """
        Retrieve only those sources from the catalog that have **not** been
        matched to the sources in the reference catalog.

        """
        mask = self._catalog['matched_ref_id'].mask
        return self._catalog[mask]

    def get_matched_cat(self):
        """
        Retrieve only those sources from the catalog that **have been**
        matched to the sources in the reference catalog.

        """
        mask = np.logical_not(self._catalog['matched_ref_id'].mask)
        return self._catalog[mask]

    def recalc_catalog_radec(self):
        """ Recalculate RA and DEC of the sources in the catalog.
        """
        for k, image in enumerate(self._images):

            idx = (self._catalog['_imcat_idx'] == k)
            if not np.any(idx):
                continue

            ra, dec = image.det_to_world(
                self._catalog['x'][idx], self._catalog['y'][idx]
            )
            self._catalog['RA'][idx] = ra
            self._catalog['DEC'][idx] = dec

    def calc_tanp_xy(self, tanplane_wcs):
        """
        Compute x- and y-positions of the sources from the image catalog
        in the tangent plane. This creates the following
        columns in the catalog's table: ``'TPx'`` and ``'TPy'``.

        Parameters
        ----------
        tanplane_wcs: TPWCS
            A `TPWCS` object that will provide transformations to
            the tangent plane to which sources of this catalog a should be
            "projected".

        """
        if 'RA' not in self._catalog.colnames or \
           'DEC' not in self._catalog.colnames:
            raise RuntimeError("'recalc_catalog_radec()' should have been run "
                               "prior to calc_tanp_xy()")

        # compute x & y in the reference WCS:
        xtp, ytp = tanplane_wcs.world_to_tanp(self.catalog['RA'],
                                              self.catalog['DEC'])
        self._catalog['TPx'] = table.MaskedColumn(
            xtp, name='TPx', dtype=np.double, mask=False
        )
        self._catalog['TPy'] = table.MaskedColumn(
            ytp, name='TPy', dtype=np.double, mask=False
        )

    def match2ref(self, refcat, match=None):
        """ Uses ``xyxymatch`` to cross-match sources between this catalog and
            a reference catalog.

        Parameters
        ----------
        refcat: RefCatalog
            A `RefCatalog` object that contains a catalog of reference sources
            as well as a valid reference WCS.

        match: MatchCatalogs, function, None, optional
            A callable that takes two arguments: a reference catalog and an
            image catalog. Both catalogs will have columns ``'TPx'`` and
            ``'TPy'`` that represent the source coordinates in some common
            (to both catalogs) coordinate system.

        """
        colnames = self._catalog.colnames
        catlen = len(self._catalog)

        if match is None:
            if catlen != len(refcat.catalog):
                raise ValueError("When matching is not requested, catalogs "
                                 "should have been matched previously and "
                                 "have equal lengths.")
            log.info("No matching of sources from '{:}' catalog with sources "
                     "from the reference '{:}' catalog was requested."
                     .format(self.name, refcat.name))
            log.info("Catalogs are assumed matched with 1-to-1 "
                     "correspondence.")
            mref_idx = np.arange(catlen)
            minput_idx = np.arange(catlen)
            nmatches = catlen

        else:
            if 'TPx' not in colnames or 'TPy' not in colnames:
                raise RuntimeError("'calc_tanp_xy()' should have been run "
                                   "prior to match2ref()")

            mref_idx, minput_idx = match(refcat.catalog, self._catalog)
            nmatches = len(mref_idx)

        # matched_ref_id:
        if 'matched_ref_id' in colnames:
            self._catalog['matched_ref_id'].mask[:] = True
        else:
            c = table.MaskedColumn(name='matched_ref_id', dtype=int,
                                   length=catlen, mask=True)
            self._catalog.add_column(c)

        self._catalog['matched_ref_id'].mask[minput_idx] = False
        self._catalog['matched_ref_id'][minput_idx] = refcat.catalog['id'][mref_idx]  # noqa: E501

        # this is needed to index reference catalog directly without using
        # astropy table indexing which, at this moment, is experimental:
        if '_raw_matched_ref_idx' in colnames:
            self._catalog['_raw_matched_ref_idx'].mask = True
        else:
            c = table.MaskedColumn(name='_raw_matched_ref_idx',
                                   dtype=int, length=catlen, mask=True)
            self._catalog.add_column(c)
        self._catalog['_raw_matched_ref_idx'][minput_idx] = mref_idx
        self._catalog['_raw_matched_ref_idx'].mask[minput_idx] = False

        log.info("Found {:d} matches for '{}'...".format(nmatches, self.name))

        # TODO: revisit this once the bug described in
        # https://github.com/spacetelescope/stsci.stimage/issues/8
        # is fixed.
        #
        # Due to this bug minput_idx may contain duplicate values.
        # Because of this, the above logic for masking, saving ids, and indices
        # does not work reliably. As a workaround, we save matched array
        # indices within the image catalog so that we can use them in
        # `fit2ref()`. For this to work, reference and image catalog must not
        # change between this function return and `fit2ref()` call.
        #
        self._mref_idx = mref_idx
        self._minput_idx = minput_idx

        return nmatches, mref_idx, minput_idx

    def fit2ref(self, refcat, tanplane_wcs, fitgeom='general', nclip=3,
                sigma=(3.0, 'rmse')):
        """
        Perform linear fit of this group's combined catalog to the reference
        catalog. When either/both group's catalog or/and the reference catalog
        contain ``'weight'`` column, weigted fitting will be performed.
        See ``Notes`` section for further details.

        Parameters
        ----------
        refcat: RefCatalog
            A `RefCatalog` object that contains a catalog of reference sources.

        tanplane_wcs: TPWCS
            A `TPWCS` object that will provide transformations to
            the tangent plane to which sources of this catalog a should be
            "projected".

        fitgeom: {'shift', 'rscale', 'general'}, optional
            The fitting geometry to be used in fitting the matched object
            lists. This parameter is used in fitting the offsets, rotations
            and/or scale changes from the matched object lists. The 'general'
            fit geometry allows for independent scale and rotation for
            each axis.

        nclip: int, None, optional
            Number (a non-negative integer) of clipping iterations in fit.
            Clipping will be turned off if ``nclip`` is either `None` or 0.

        sigma: float, tuple of the form (float, str), optional
            When a tuple is provided, first value (a positive number)
            indicates the number of "fit error estimates" to use for clipping.
            The second value (a string) indicates the statistic to be
            used for "fit error estimate". Currently the following values are
            supported: ``'rmse'``, ``'mae'``, and ``'std'``
            - see `~tweakwcs.linearfit.iter_linear_fit` for more details.

            When ``sigma`` is a single number, it must be a positive number and
            the default error estimate ``'rmse'`` is assumed.

            This parameter is ignored when ``nclip`` is either `None` or 0.

        Notes
        -----
        When fitting image sources to reference catalog sources, we can
        specify which sources have higher weights. This can be done by
        assigning a "weight" to each source by specifying these values
        in the optional ``'weight'`` column of either the reference catalog,
        image catalog, or both.

        When weights are not provided, all sources are weighed equally. When
        only one of image or reference catalog weights are provided,
        the sources will be weighted with the specified weights.
        When *both* image *and* reference catalogs specify weights for
        the same sources, the two weights will be combined into a single
        weight as:

        .. math::
            1/w = 1/w_i + 1/w_r

        .. warning::
            Keep in mind that when a group catalog is created from individual
            catalogs, weights of the group catalog are created by
            *concatenating* weights of individual catalogs. Therefore,
            for the weighting of groups of catalogs to work correctly,
            the weights of individual catalogs should be scaled in such a way
            that when individual catalogs are combined into a single
            "group catalog", weights preserve their relative values.

            For example, let's say a group is formed from two individual
            catalogs. Let's say first catalog contains four sources with equal
            weights ``[1,1,1,1]`` and the second catalog contains two sources
            with weights ``[1,1]`` then the group's catalogs sources will
            also have equal weights ``[1,1,1,1,1,1]``. However, if each
            individual catalog's weights were normalized such that sum of
            all weights is 1, then group's sources will be weighed unequally:
            ``[0.25,0.25,0.25,0.25,0.5,0.5]``.

        """
        im_xyref = np.asarray([self._catalog['TPx'],
                               self._catalog['TPy']]).T
        refxy = np.asarray([refcat.catalog['TPx'],
                            refcat.catalog['TPy']]).T

        # mask = np.logical_not(self._catalog['matched_ref_id'].mask)
        # im_xyref = im_xyref[mask]
        # ref_idx = self._catalog['_raw_matched_ref_idx'][mask]

        # TODO: revisit this once the bug described in
        # https://github.com/spacetelescope/stsci.stimage/issues/8
        # is fixed.
        #
        # Due to this bug minput_idx may contain duplicate values.
        # For now we bypass the above commented code by accessing indices
        # stored in the image's catalog.
        minput_idx = self._minput_idx
        im_xyref = im_xyref[minput_idx]
        ref_idx = self._mref_idx

        refxy = refxy[ref_idx]

        # process weights:
        if 'weight' in self._catalog.colnames:
            im_weight = np.asarray(self._catalog['weight'])[minput_idx]
        else:
            im_weight = None

        if 'weight' in refcat.catalog.colnames:
            ref_weight = np.asarray(refcat.catalog['weight'])[ref_idx]
        else:
            ref_weight = None

        fit = iter_linear_fit(
            refxy, im_xyref, ref_weight, im_weight,
            fitgeom=fitgeom, nclip=nclip, sigma=sigma, center=None
        )

        # re-compute shifts for the center at (0, 0):
        fit['shift_ld'] += fit['center_ld'] - np.dot(fit['center_ld'], fit['matrix_ld'].T)
        fit['shift'] = fit['shift_ld'].astype(np.double)

        xy_fit = fit['shift'] + np.dot(im_xyref[fit['fitmask']], fit['matrix'].T)

        fit['fit_xy'] = xy_fit
        fit['fit_RA'], fit['fit_DEC'] = tanplane_wcs.tanp_to_world(*(xy_fit.T))

        log.info("Computed '{:s}' fit for {}:".format(fitgeom, self.name))
        if fitgeom == 'shift':
            log.info("XSH: {:.6g}  YSH: {:.6g}".format(*fit['shift']))
        elif fitgeom in ['rshift', 'rscale'] and fit['proper']:
            log.info(
                "XSH: {:.6g}  YSH: {:.6g}    ROT: {:.6g}    SCALE: {:.6g}"
                .format(*fit['shift'], fit['proper_rot'], fit['<scale>'])
            )
        elif fitgeom == 'general' or (fitgeom in ['rshift', 'rscale'] and not
                                      fit['proper']):
            log.info("XSH: {:.6g}  YSH: {:.6g}    PROPER ROT: {:.6g}    "
                     .format(*fit['shift'], fit['proper_rot']))
            log.info("<ROT>: {:.6g}  SKEW: {:.6g}    ROT_X: {:.6g}  "
                     "ROT_Y: {:.6g}".format(fit['<rot>'], fit['skew'],
                                            *fit['rot']))
            log.info("<SCALE>: {:.6g}  SCALE_X: {:.6g}  SCALE_Y: {:.6g}"
                     .format(fit['<scale>'], *fit['scale']))
        else:
            raise ValueError("Unsupported fit geometry.")  # pragma: no cover

        log.info("")
        log.info("FIT RMSE: {:.6g}   FIT MAE: {:.6g}"
                 .format(fit['rmse'], fit['mae']))
        log.info("Final solution based on {:d} objects."
                 .format(fit['resids'].shape[0]))

        return fit

    @deprecated_renamed_argument('tanplane_wcs', 'ref_tpwcs', since='0.6.5')
    def apply_affine_to_wcs(self, ref_tpwcs, matrix, shift, meta=None):
        """ Applies a general affine transformation to the WCS. """
        for imcat in self:
            imcat.tpwcs.set_correction(matrix, shift, ref_tpwcs=ref_tpwcs, meta=meta)

    def align_to_ref(self, refcat, ref_tpwcs=None, match=None, minobj=None,
                     fitgeom='rscale', nclip=3, sigma=(3.0, 'rmse')):
        """
        Matches sources from the image catalog to the sources in the
        reference catalog, finds the affine transformation between matched
        sources, and adjusts images' WCS according to this fit.

        Upon successful return, this function will also set the following
        fields of the ``fit_info`` attribute of each member
        `WCSImageCatalog` object:

            * **'fitgeom'**: the value of the ``fitgeom`` argument
            * **'eff_minobj'**: effective value of the ``minobj`` parameter
            * **'matrix'**: computed rotation matrix
            * **'shift'**: shift (offset) along X- and Y-axis
            * **'rot'**: A tuple of ``(rotx, roty)`` - the rotation angles with
              regard to the ``X`` and ``Y`` axes.
            * **'<rot>'**: *Arithmetic mean* of the angles of rotation around
              ``X`` and ``Y`` axes.
            * **'proper_rot'**: rotation angle as if rotation is a proper
              rotation.
            * **'proper'**: Indicates whether the rotation is a proper rotation
              (boolean)
            * **'scale'**: A tuple of ``(sx, sy)`` - scale change in the
              direction of the ``X`` and ``Y`` axes.
            * **'<scale>'**: *Geometric mean* of scales ``sx`` and ``sy``.
            * **'skew'**: Computed skew - an angle in the range [-180, 180).
            * **'center'**: Center of rotation in the *tangent plane* of the
              computed linear transformations.
            * **'fitmask'**: boolean array indicating (with `True`) sources
              **used** for fitting
            * **'nmatches'** [when ``match`` is not `None`]: number of matched
              sources
            * **'matched_ref_idx'** [when ``match`` is not `None`]: indices of
              the matched sources in the reference catalog
            * **'matched_input_idx'** [when ``match`` is not `None`]: indices
              of the matched sources in the "input" catalog (the catalog from
              image to be aligned)
            * **'fit_ref_idx'**: indices of the sources from the reference
              catalog used for fitting (a subset of 'matched_ref_idx' indices,
              when ``match`` is not `None`, left after clipping iterations
              performed during fitting)
            * **'fit_input_idx'**: indices of the sources from the "input"
              (image) catalog used for fitting (a subset of
              'matched_input_idx' indices, when ``match`` is not `None`,
              left after clipping iterations performed during fitting)
            * **'rmse'**: fit Root-Mean-Square Error in *tangent plane*
              coordinates of corrected image source positions from reference
              source positions.
            * **'mae'**: fit Mean Absolute Error in *tangent plane*
              coordinates of corrected image source positions from reference
              source positions.
            * **'std'**: Norm of the STandard Deviation of the residuals
              in *tangent plane* along each axis.
            * **'fit_RA'**: first (corrected) world coordinate of input source
              positions used in fitting.
            * **'fit_DEC'**: second (corrected) world coordinate of input
              source positions used in fitting.
            * **'status'**: Alignment status. Currently two possible status are
              possible ``'SUCCESS'`` or ``'FAILED: reason for failure'``.
              When alignment failed, the reason for failure is provided after
              alignment status.

        .. note::
            A ``'SUCCESS'`` status does not indicate a "good" alignment. It
            simply indicates that alignment algortithm has completed without
            errors. Use other fields to evaluate alignment: fit ``RMSE``
            and ``MAE`` values, number of matched sources, etc.

        Parameters
        ----------
        refcat: RefCatalog
            A `RefCatalog` object that contains a catalog of reference sources.

        ref_tpwcs : TPWCS
            A `TPWCS` object that defines a projection tangent plane to be
            used for matching and fitting during alignment.

        match: MatchCatalogs, function, None, optional
            A callable that takes two arguments: a reference catalog and an
            image catalog.

        minobj: int, None, optional
            Minimum number of identified objects from each input image to use
            in matching objects from other images. If the default `None` value
            is used then `align` will automatically deternmine the minimum
            number of sources from the value of the `fitgeom` parameter.
            This parameter is used to interrupt alignment process (catalog
            fitting, ``WCS`` "tweaking") when the number of matched sources
            is smaller than the value of ``minobj`` in which case this
            method will return `False`.

        fitgeom: {'shift', 'rscale', 'general'}, optional
            The fitting geometry to be used in fitting the matched object
            lists. This parameter is used in fitting the offsets, rotations
            and/or scale changes from the matched object lists. The 'general'
            fit geometry allows for independent scale and rotation for each
            axis. This parameter is ignored if ``match`` is `False`.

        nclip: int, None, optional
            Number (a non-negative integer) of clipping iterations in fit.
            Clipping will be turned off if ``nclip`` is either `None` or 0.

            This parameter is ignored if ``match`` is `False`.

        sigma: float, tuple of the form (float, str), optional
            When a tuple is provided, first value (a positive number)
            indicates the number of "fit error estimates" to use for clipping.
            The second value (a string) indicates the statistic to be
            used for "fit error estimate". Currently the following values are
            supported: ``'rmse'``, ``'mae'``, and ``'std'``
            - see `~tweakwcs.linearfit.iter_linear_fit` for more details.

            When ``sigma`` is a single number, it must be a positive number and
            the default error estimate ``'rmse'`` is assumed.

            This parameter is ignored when ``nclip`` is either `None` or 0
            or when ``match`` is `False`.

        Returns
        -------
        bool
            Returns `True` if the number of matched sources is larger or equal
            to ``minobj`` and all steps have been performed, including catalog
            fitting and ``WCS`` "tweaking". If the number of matched sources is
            smaller than ``minobj``, this function will return `False`.

        """
        if not self._images:
            name = 'Unnamed' if self.name is None else self.name
            log.warning("WCSGroupCatalog '{:s}' is empty. Nothing to align."
                        .format(name))
            return False

        # set initial status to 'FAILED':
        for imcat in self:
            imcat.fit_status = "FAILED: Unknown error"

        if minobj is None:
            minobj = SUPPORTED_FITGEOM_MODES[fitgeom]
            log.debug(f"Setting 'minobj' to {minobj} for fitgeom='{fitgeom}'")

        if ref_tpwcs is None:
            ref_tpwcs = deepcopy(self._images[0].tpwcs)

        self.calc_tanp_xy(tanplane_wcs=ref_tpwcs)
        refcat.calc_tanp_xy(tanplane_wcs=ref_tpwcs)

        nmatches, mref_idx, minput_idx = self.match2ref(
            refcat=refcat,
            match=match
        )

        if nmatches < minobj:
            name = 'Unnamed' if self.name is None else self.name
            log.warning("Not enough matches (< {:d}) found for image "
                        "catalog '{:s}'.".format(nmatches, name))
            for imcat in self:
                imcat.fit_status = 'FAILED: not enough matches'
            return False

        fit = self.fit2ref(refcat=refcat, tanplane_wcs=ref_tpwcs,
                           fitgeom=fitgeom, nclip=nclip, sigma=sigma)

        fit_info = {
            'fitgeom': fitgeom,
            'eff_minobj': minobj,
            'matrix': fit['matrix'],
            'shift': fit['shift'],
            'center': fit['center'],  # center of rotation in geom. transforms
            'fitmask': fit['fitmask'],  # sources was used for fitting
            'proper_rot': fit['proper_rot'],  # proper rotation
            'proper': fit['proper'],  # is a proper rotation? True/False
            'rot': fit['rot'],  # rotx, roty
            '<rot>': fit['<rot>'],  # Arithmetic mean of rotx and roty
            'scale': fit['scale'],  # sx, sy
            '<scale>': fit['<scale>'],  # Geometric mean of sx, sy
            'skew': fit['skew'],  # skew
            'rmse': fit['rmse'],  # fit RMSE in tangent plane coords
            'mae': fit['mae'],  # fit MAE in tangent plane coords
            'fit_RA': fit['fit_RA'],
            'fit_DEC': fit['fit_DEC'],
            'status': 'SUCCESS',
        }

        if match is not None:
            fit_info.update({
                'nmatches': nmatches,
                'matched_ref_idx': mref_idx,
                'matched_input_idx': minput_idx
            })

        self.apply_affine_to_wcs(
            ref_tpwcs=ref_tpwcs,
            matrix=fit['matrix'],
            shift=fit['shift'],
            # meta=meta
        )

        for imcat in self:
            imcat.fit_info.update(deepcopy(fit_info))

        self.recalc_catalog_radec()

        return True


class RefCatalog(object):
    """
    An object that holds a reference catalog and provides
    tools for coordinate convertions using reference WCS as well as
    catalog manipulation and expansion.

    """

    def __init__(self, catalog, name=None, footprint_tol=1.0):
        """
        Parameters
        ----------
        catalog: astropy.table.Table
            Reference catalog.

            .. note::
                Reference catalogs (:py:class:`~astropy.table.Table`)
                *must* contain *both* ``'RA'`` and ``'DEC'`` columns.

        name: str, None, optional
            Name of the reference catalog.

        footprint_tol: float, optional
            Matching tolerance in arcsec. This is used to estimate catalog's
            footprint when catalog contains only one or two sources.

        """
        self._name = name
        self._catalog = None
        self._footprint_tol = footprint_tol
        self._poly_area = None

        # make sure catalog has RA & DEC
        if catalog is not None:
            self.catalog = catalog

    def _check_catalog(self, catalog):
        if catalog is None:
            raise ValueError("Reference catalogs cannot be None")

        if 'RA' not in catalog.colnames or 'DEC' not in catalog.colnames:
            raise KeyError("Reference catalogs *must* contain *both* 'RA' "
                           "and 'DEC' columns.")

    @property
    def name(self):
        """ Get/set :py:class:`RefCatalog` object's name.
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def catalog(self):
        """ Get/set image's catalog.
        """
        return self._catalog

    @catalog.setter
    def catalog(self, catalog):
        self._check_catalog(catalog)

        if not catalog:
            raise ValueError("Reference catalog must contain at least one "
                             "source.")

        self._catalog = catalog.copy()

        if 'id' not in self._catalog.colnames:
            self._catalog['id'] = np.arange(1, len(self._catalog) + 1)

        # create spherical polygon bounding the sources
        self.calc_bounding_polygon()

    @property
    def poly_area(self):
        """ Area of the bounding polygon (in srad).
        """
        return self._poly_area

    @property
    def polygon(self):
        """ Get image's (or catalog's) bounding spherical polygon.
        """
        return self._polygon

    def intersection(self, wcsim):
        """
        Compute intersection of this `WCSImageCatalog` object and another
        `WCSImageCatalog`, `WCSGroupCatalog`, `RefCatalog`, or
        :py:class:`~spherical_geometry.polygon.SphericalPolygon`
        object.

        Parameters
        ----------
        wcsim: WCSImageCatalog, WCSGroupCatalog, RefCatalog, SphericalPolygon
            Another object that should be intersected with this
            `WCSImageCatalog`.

        Returns
        -------
        polygon: SphericalPolygon
            A :py:class:`~spherical_geometry.polygon.SphericalPolygon` that is
            the intersection of this `WCSImageCatalog` and `wcsim`.

        """
        if isinstance(wcsim, (WCSImageCatalog, WCSGroupCatalog, RefCatalog)):
            return self._polygon.intersection(wcsim.polygon)
        else:
            return self._polygon.intersection(wcsim)

    # TODO: due to a bug in the sphere package, see
    #       https://github.com/spacetelescope/sphere/issues/74
    #       intersections with polygons formed as union does not work.
    #       For this reason I re-implement 'intersection_area' below with
    #       a workaround for the bug.
    #       The original implementation should be uncommented once the bug
    #       is fixed.
    #
    # def intersection_area(self, wcsim):
        # """ Calculate the area of the intersection polygon.
        # """
        # return np.fabs(self.intersection(wcsim).area())
    def intersection_area(self, wcsim):
        """ Calculate the area of the intersection polygon.
        """
        if isinstance(wcsim, (WCSImageCatalog, RefCatalog)):
            return np.fabs(self.intersection(wcsim).area())

        else:
            # this is bug workaround:
            area = 0.0
            for wim in wcsim:
                area += np.fabs(
                    self.polygon.intersection(wim.polygon).area()
                )
            return area

    def _guarded_intersection_area(self, wcsim):
        """
        Calculate the area of the intersection polygon. If some
        intersections fail due to a bug/limitation of ``spherical_geometry``
        then the area of the valid intersections will be returned.
        If images do not intersect or intersection fails, 0 will be returned.
        """
        if isinstance(wcsim, (WCSImageCatalog, RefCatalog)):
            try:
                return np.fabs(self.intersection(wcsim).area())
            except MalformedPolygonError:
                return 0.0

        else:
            # this is bug workaround:
            area = 0.0
            for wim in wcsim:
                try:
                    area += np.fabs(
                        self.polygon.intersection(wim.polygon).area()
                    )
                except MalformedPolygonError:
                    continue
            return area

    def _calc_cat_convex_hull(self):
        """
        Calculate spherical polygon corresponding to the convex hull bounding
        the sources in the catalog.

        """
        if self.catalog is None:
            return

        # Find an "optimal" tangent plane to the catalog points based on the
        # mean point and then construct a WCS based on the mean point.
        # Compute x, y coordinates in this tangent plane based on the
        # previously computed WCS and return the set of x, y coordinates and
        # "reference WCS".
        x, y, z = _S2C(self.catalog['RA'], self.catalog['DEC'])
        ra_ref, dec_ref = _C2S(
            x.mean(dtype=np.double),
            y.mean(dtype=np.double),
            z.mean(dtype=np.double)
        )

        rotm = [planar_rot_3d(np.deg2rad(alpha), 2 - axis)
                for axis, alpha in enumerate([ra_ref, dec_ref])]
        euler_rot = np.linalg.multi_dot(rotm)
        inv_euler_rot = inv(euler_rot)
        xr, yr, zr = np.dot(euler_rot, (x, y, z))
        x = yr / xr
        y = zr / xr

        xv, yv = convex_hull(x, y)

        if len(xv) == 0:
            # no points
            raise RuntimeError(  # pragma: no cover
                "Unexpected error: Contact software developer"
            )

        elif len(xv) == 1:
            # one point. build a small box around it:
            x, y = convex_hull(x, y, wcs=None)
            tol = 0.5 * self._footprint_tol

            xv = [x[0] - tol, x[0] - tol, x[0] + tol, x[0] + tol, x[0] - tol]
            yv = [y[0] - tol, y[0] + tol, y[0] + tol, y[0] - tol, y[0] - tol]

        elif len(xv) == 2 or len(xv) == 3:
            # two points. build a small box around them:
            x, y = convex_hull(x, y, wcs=None)
            tol = 0.5 * self._footprint_tol

            vx = y[1] - y[0]
            vy = x[1] - x[0]
            norm = np.sqrt(vx * vx + vy * vy)
            vx /= norm
            vy /= norm

            xv = [
                x[0] - (vx - vy) * tol,
                x[0] - (vx + vy) * tol,
                x[1] + (vx - vy) * tol,
                x[1] + (vx + vy) * tol,
                x[0] - (vx - vy) * tol
            ]
            yv = [
                y[0] - (vy + vx) * tol,
                y[0] - (vy - vx) * tol,
                y[1] + (vy + vx) * tol,
                y[1] + (vy - vx) * tol,
                y[0] - (vy + vx) * tol
            ]

        # "unrotate" cartezian coordinates back to their original
        # ra_ref and dec_ref "positions":
        xt = np.ones_like(xv)
        xcr, ycr, zcr = np.dot(inv_euler_rot, (xt, xv, yv)).astype(np.double)
        # convert cartesian to spherical coordinates:
        ra, dec = _C2S(xcr, ycr, zcr)

        # TODO: for strange reasons, occasionally ra[0] != ra[-1] and/or
        #       dec[0] != dec[-1] (even though we close the polygon in the
        #       previous two lines). Then SphericalPolygon fails because
        #       points are not closed. Threfore we force it to be closed:
        ra[-1] = ra[0]
        dec[-1] = dec[0]

        self._radec = [(ra, dec)]
        self._polygon = SphericalPolygon.from_radec(ra, dec)
        self._poly_area = np.fabs(self._polygon.area())

    def calc_bounding_polygon(self):
        """ Calculate bounding polygon of the sources in the catalog.
        """
        # create spherical polygon bounding the sources
        self._calc_cat_convex_hull()

    def expand_catalog(self, catalog):
        """
        Expand current reference catalog with sources from another catalog.

        If current catalog is empty, then the catalog being added will become
        the new reference catalog. In this case if the ``catalog`` does have
        ``id`` column, those ID values will be preserved. If the ``catalog``
        does not contain an ID column, then the new IDs will be assigned
        in increasing order starting with 1.

        If the existing reference catalog is not empty, then the IDs from the
        ``catalog`` being added will be discarded and new IDs will be assigned
        in the increasing order such as to continue the numbering of existing
        source positions in the reference catalog.

        Parameters
        ----------
        catalog: astropy.table.Table
            A catalog of new sources to be added to the existing reference
            catalog. `catalog` *must* contain *both* ``'RA'`` and ``'DEC'``
            columns.

        """
        self._check_catalog(catalog)
        cat = catalog.copy()
        if self._catalog is None:
            self._catalog = cat
            if 'id' not in self._catalog.colnames:  # pragma: no branch
                self._catalog['id'] = np.arange(1, len(self._catalog) + 1)
        else:
            maxid = self.catalog['id'].max()
            oldlen = len(self.catalog)
            self._catalog = table.vstack(
                [self.catalog, cat],
                join_type='outer',
                metadata_conflicts='silent'
            )

            # assign ids to the newly added source positions in consecutive
            # order above the largest id in the already existing catalog:
            self._catalog['id'][oldlen:] = np.arange(maxid + 1,
                                                     maxid + len(cat) + 1)

        self.calc_bounding_polygon()

    def calc_tanp_xy(self, tanplane_wcs):
        """
        Compute x- and y-positions of the sources from the reference catalog
        in the tangent plane provided by the `tanplane_wcs`.
        This creates the following columns in the catalog's table:
        ``'TPx'`` and ``'TPy'``.

        Parameters
        ----------
        tanplane_wcs: TPWCS
            A `TPWCS` object that will provide transformations to
            the tangent plane to which sources of this catalog a should be
            "projected".

        """
        # compute x & y in the reference WCS:
        xtp, ytp = tanplane_wcs.world_to_tanp(self.catalog['RA'],
                                              self.catalog['DEC'])
        self._catalog['TPx'] = table.MaskedColumn(
            xtp, name='TPx', dtype=np.double, mask=False
        )
        self._catalog['TPy'] = table.MaskedColumn(
            ytp, name='TPy', dtype=np.double, mask=False
        )


def convex_hull(x, y, wcs=None):
    """Computes the convex hull of a set of 2D points.

    Implements `Andrew's monotone chain algorithm <http://en.wikibooks.org\
/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain>`_.
    The algorithm has O(n log n) complexity.

    Credit: `<http://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/\
Convex_hull/Monotone_chain>`_

    Parameters
    ----------

    points: list of tuples
        An iterable sequence of (x, y) pairs representing the points.

    Returns
    -------
    Output: list
        A list of vertices of the convex hull in counter-clockwise order,
        starting from the vertex with the lexicographically smallest
        coordinates.

    """
    ndarray = isinstance(x, np.ndarray) or isinstance(y, np.ndarray)

    # Sort the points lexicographically (tuples are compared
    # lexicographically). Remove duplicates to detect the case we have
    # just one unique point.
    points = sorted(set(zip(x, y)))

    # Boring case: no points or a single point,
    # possibly repeated multiple times.
    if len(points) == 0:
        if ndarray:
            return (np.array([]), np.array([]))
        else:
            return ([], [])

    elif len(points) == 1:
        if ndarray:
            return (np.array([points[0][0]]), np.array([points[0][1]]))
        else:
            return ([points[0][0]], [points[0][1]])

    # 2D cross product of OA and OB vectors, i.e. z-component of their
    # 3D cross product.
    # Returns a positive value, if OAB makes a counter-clockwise turn,
    # negative for clockwise turn, and zero if the points are collinear.
    def cross(o, a, b):
        return (a[0] - o[0]) * (b[1] - o[1]) - (a[1] - o[1]) * (b[0] - o[0])

    # Build lower hull
    lower = []
    for p in points:
        while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
            lower.pop()
        lower.append(p)

    # Build upper hull
    upper = []
    for p in reversed(points):
        while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
            upper.pop()
        upper.append(p)

    # Concatenation of the lower and upper hulls gives the convex hull.
    # Last point of each list is omitted because it is repeated at the
    # beginning of the other list.
    total_hull = np.asanyarray(lower[:-1] + upper)

    ptx = total_hull[:, 0]
    pty = total_hull[:, 1]

    if wcs is None:
        if ndarray:
            return (ptx, pty)
        else:
            return (ptx.tolist(), pty.tolist())

    # convert x, y vertex coordinates to RA & DEC:
    ra, dec = wcs(ptx, pty)
    ra[-1] = ra[0]
    dec[-1] = dec[0]

    if ndarray:
        return (ra, dec)
    else:
        return (ra.tolist(), dec.tolist())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
r"""
A module that provides functions for automatically adding ``group_id``
tag to images based on information recorded in image's ``meta`` attribute.

This ``tweakwcs_group_id`` tag, if set, can be used by ``align_wcs``,
and other functions from the ``imalign`` module to
identify images that should be treated together as a group when "tweaking"
their ``WCS``. That is, all images within a group will have the same
correction applied to their ``WCS``\ es. This is often the case with images
that come from differenct chips of the same sensor chip assembly (SCA).

.. note::
    Grouping logic/algorithms are inherently telescope and instrument
    dependent.

:Authors: Mihai Cara (contact: help@stsci.edu)

:License: :doc:`../LICENSE`

"""
# STDLIB
import logging
import uuid

# LOCAL
from .. import __version__  # noqa: F401


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def assign_jwst_tweakwcs_groups(images):
    """ Assign group IDs to ``JWST`` images.

    Parameters
    ----------
    images: list of str or jwst.datamodels.DataModel
        A list of string file names to ``JWST`` data or
        `jwst.datamodels.DataModel` objects. On return, these data models
        or files will contain ``tweakwcs_group_id`` attribute in their
        ``meta`` dictionary identifying the tweakwcs group. Images within
        a group are aligned together.

    """
    try:
        from jwst.datamodels import open as open_data  # pylint: disable=W0611
    except ImportError:
        raise ImportError("'assign_jwst_tweakwcs_groups' requires that "
                          "'jwst' package be installed.")

    close = [isinstance(im, (str, bytes)) for im in images]

    group_ids = {}

    for im, cl in zip(images, close):
        model = open_data(im) if cl else im
        try:
            meta_ids = (
                model.meta.observation.program_number,
                model.meta.observation.observation_number,
                model.meta.observation.visit_number,
                model.meta.observation.visit_group,
                model.meta.observation.sequence_id,
                model.meta.observation.activity_id,
                model.meta.observation.exposure_number
            )

            if meta_ids in group_ids:
                gid = group_ids[meta_ids]
            else:
                gid = str(uuid.uuid4())
                # next loop will almost never execute as UUIDs should be
                # unique.
                while gid in group_ids.values():
                    gid = str(uuid.uuid4())  # pragma: no cover
                group_ids[meta_ids] = gid

        except Exception as e:
            gid = 'None'
            log.warning("Unable to assign a 'tweakwcs_group_id' to image "
                        "'{}' due to '{}'".format(model.meta['filename'], e))
            log.warning("'tweakwcs_group_id' for image '{}' will be set "
                        "to None".format(model.meta['filename']))

        finally:
            if "tweakwcs_group_id" not in model.meta:  # pragma: no branch
                twgid_schema = {
                    "type": "object",
                    "properties": {
                        "meta": {
                            "type": "object",
                            "properties": {
                                "tweakwcs_group_id": {
                                    "title": "tweakwcs group ID",
                                    "type": "string",
                                    "fits_keyword": "TWEAKGID"
                                }
                            }
                        }
                    }
                }
                model.extend_schema(twgid_schema)
            model.meta.tweakwcs_group_id = gid

            if cl:
                model.save(model.meta.filename)



"""
This package provides support for image alignment.

"""
__docformat__ = 'restructuredtext'

__taskname__ = 'tweakwcs'
__author__ = 'Mihai Cara'

from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = 'UNKNOWN'

from .tpwcs import TPWCS, JWSTgWCS, FITSWCS  # noqa: F401
from .matchutils import MatchCatalogs, TPMatch  # noqa: F401
from .imalign import fit_wcs, align_wcs  # noqa: F401
from .wcsimage import (convex_hull, RefCatalog, WCSImageCatalog,  # noqa: F401
                       WCSGroupCatalog)  # noqa: F401
from .linalg import inv  # noqa: F401
from .linearfit import iter_linear_fit, build_fit_matrix  # noqa: F401



"""
A module containing unit tests for the `wcsutil` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
import pytest
from distutils.version import LooseVersion

import numpy as np

from tweakwcs import wcsutils

try:
    import gwcs
    if LooseVersion(gwcs.__version__) > '0.12.0':
        from gwcs.geometry import SphericalToCartesian, CartesianToSpherical
        _S2C = SphericalToCartesian(name='s2c', wrap_lon_at=180)
        _C2S = CartesianToSpherical(name='c2s', wrap_lon_at=180)
        _GWCS_VER_GT_0P12 = True
    else:
        _GWCS_VER_GT_0P12 = False
except ImportError:
    _GWCS_VER_GT_0P12 = False

_NO_JWST_SUPPORT = not _GWCS_VER_GT_0P12


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
@pytest.mark.parametrize('x,y,z', [
    (1, 0, 0),
    (0, 1, 0),
    (0, 0, 1),
    (-1, 0, 0),
    (0, -1, 0),
    (0, 0, -1),
])
def test_cartesian_spherical_cartesian_roundtrip_special(x, y, z):
    feps = 100 * np.finfo(np.double).eps
    xyz = _S2C(*_C2S(x, y, z))
    assert np.allclose((x, y, z), xyz, rtol=0, atol=feps)


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_cartesian_spherical_cartesian_roundtrip_rand():
    feps = 100 * np.finfo(np.double).eps
    xyz = np.random.random((100, 3))
    xyz /= np.linalg.norm(xyz, axis=1)[:, np.newaxis]
    x, y, z = xyz.T
    rx, ry, rz = _S2C(*_C2S(x, y, z))
    assert np.allclose(rx, x, rtol=0, atol=feps)
    assert np.allclose(ry, y, rtol=0, atol=feps)
    assert np.allclose(rz, z, rtol=0, atol=feps)


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_spherical_cartesian_spherical_roundtrip_ugrid():
    feps = 1000 * np.finfo(np.double).eps
    angles = np.linspace(-180, 180, 13)
    alpha0 = np.repeat(angles, angles.size)
    delta0 = np.tile(angles / 2, angles.size)
    alpha, delta = _C2S(*_S2C(alpha0, delta0))
    assert np.allclose(alpha, alpha0, rtol=0, atol=feps)
    assert np.allclose(delta, delta0, rtol=0, atol=feps)


@pytest.mark.parametrize('angle', np.linspace(-2 * np.pi, 2 * np.pi, 100))
def test_planar_rot_3d(angle):
    feps = 100 * np.finfo(np.double).eps
    ref = ((angle - np.pi) % (2 * np.pi)) - np.pi
    for axis in (0, 1, 2, 1.0, 2.0, -0.0):
        mat_2d = np.delete(
            np.delete(
                wcsutils.planar_rot_3d(angle=angle, axis=axis), int(axis), 0
            ), int(axis), 1
        )
        assert np.allclose(np.linalg.det(mat_2d), 1, rtol=0, atol=feps)
        assert np.allclose(np.arctan2(*([-1, 1] * mat_2d[:, 0][::-1])),
                           ref, rtol=feps, atol=feps)
        assert np.allclose(np.arctan2(*mat_2d[:, 1]), ref, rtol=feps,
                           atol=feps)


@pytest.mark.parametrize('axis', [-1, 3, 4.5])
def test_test_planar_rot_3d_axis_out_of_range(axis):
    with pytest.raises(ValueError):
        wcsutils.planar_rot_3d(angle=np.random.random(), axis=axis)



from docutils.nodes import figure, caption, Text, reference, raw, SkipNode
from sphinx.roles import XRefRole


# Element classes

class page_ref(reference):
    pass


class num_ref(reference):
    pass


# Visit/depart functions

def skip_page_ref(self, node):
    raise SkipNode


def latex_visit_page_ref(self, node):
    self.body.append("\\pageref{%s:%s}" % (node['refdoc'], node['reftarget']))
    raise SkipNode


def latex_visit_num_ref(self, node):
    fields = node['reftarget'].split('#')
    if len(fields) > 1:
        label, target = fields
        ref_link = '%s:%s' % (node['refdoc'], target)
        latex = "\\hyperref[%s]{%s \\ref*{%s}}" % (ref_link, label, ref_link)
        self.body.append(latex)
    else:
        self.body.append('\\ref{%s:%s}' % (node['refdoc'], fields[0]))

    raise SkipNode


def doctree_read(app, doctree):
    # first generate figure numbers for each figure
    env = app.builder.env
    figid_docname_map = getattr(env, 'figid_docname_map', {})

    for figure_info in doctree.traverse(figure):
        for id in figure_info['ids']:
            figid_docname_map[id] = env.docname

    env.figid_docname_map = figid_docname_map


def doctree_resolved(app, doctree, docname):
    i = 1
    figids = {}
    for figure_info in doctree.traverse(figure):
        if app.builder.name != 'latex' and app.config.number_figures:
            for cap in figure_info.traverse(caption):
                cap[0] = Text("%s %d: %s" % (app.config.figure_caption_prefix, i, cap[0]))

        for id in figure_info['ids']:
            figids[id] = i

        i += 1

    # replace numfig nodes with links
    if app.builder.name != 'latex':
        for ref_info in doctree.traverse(num_ref):
            if '#' in ref_info['reftarget']:
                label, target = ref_info['reftarget'].split('#')
                labelfmt = label + " %d"
            else:
                labelfmt = '%d'
                target = ref_info['reftarget']

            if target not in figids:
                continue

            if app.builder.name == 'html':
                target_doc = app.builder.env.figid_docname_map[target]
                link = "%s#%s" % (app.builder.get_relative_uri(docname, target_doc),
                                  target)
                html = '<a class="pageref" href="%s">%s</a>' % (link, labelfmt % (figids[target]))
                ref_info.replace_self(raw(html, html, format='html'))
            else:
                ref_info.replace_self(Text(labelfmt % (figids[target])))


def clean_env(app):
    app.builder.env.i = 1
    app.builder.env.figid_docname_map = {}


def setup(app):
    app.add_config_value('number_figures', True, True)
    app.add_config_value('figure_caption_prefix', "Figure", True)

    app.add_node(page_ref,
                 text=(skip_page_ref, None),
                 html=(skip_page_ref, None),
                 latex=(latex_visit_page_ref, None))

    app.add_role('page', XRefRole(nodeclass=page_ref))

    app.add_node(num_ref,
                 latex=(latex_visit_num_ref, None))

    app.add_role('num', XRefRole(nodeclass=num_ref))

    app.connect("builder-inited", clean_env)
    app.connect('doctree-read', doctree_read)
    app.connect('doctree-resolved', doctree_resolved)



from distutils.version import LooseVersion

import pytest
import numpy as np
import astropy
from astropy.io import fits
from astropy import table
from astropy import wcs as fitswcs
from astropy.utils.data import get_pkg_data_filename
from astropy.modeling import polynomial
from astropy.modeling.models import (
    Shift, AffineTransformation2D, Pix2Sky_TAN, RotateNative2Celestial,
    Identity, Mapping, Const1D, Scale
)
from astropy import units as u
from astropy import coordinates as coord

import tweakwcs


try:
    import gwcs
    if LooseVersion(gwcs.__version__) > '0.12.0':
        from gwcs.geometry import SphericalToCartesian, CartesianToSpherical
        from gwcs import coordinate_frames as cf
        _GWCS_VER_GT_0P12 = True
    else:
        _GWCS_VER_GT_0P12 = False
except ImportError:
    _GWCS_VER_GT_0P12 = False

_ASTROPY_VER_GE_4 = LooseVersion(astropy.__version__) >= '4.0'
_NO_JWST_SUPPORT = not (_ASTROPY_VER_GE_4 and _GWCS_VER_GT_0P12)

_ATOL = 1e3 * np.finfo(np.array([1.]).dtype).eps

_RAD2ARCSEC = 3600.0 * np.rad2deg(1.0)
_ARCSEC2RAD = 1.0 / _RAD2ARCSEC


def _make_gwcs_wcs(fits_hdr):
    hdr = fits.Header.fromfile(get_pkg_data_filename(fits_hdr))
    fw = fitswcs.WCS(hdr)

    a_order = hdr['A_ORDER']
    a_coeff = {}
    for i in range(a_order + 1):
        for j in range(a_order + 1 - i):
            key = 'A_{:d}_{:d}'.format(i, j)
            if key in hdr:
                a_coeff[key] = hdr[key]

    b_order = hdr['B_ORDER']
    b_coeff = {}
    for i in range(b_order + 1):
        for j in range(b_order + 1 - i):
            key = 'B_{:d}_{:d}'.format(i, j)
            if key in hdr:
                b_coeff[key] = hdr[key]

    distortion = polynomial.SIP(
        fw.wcs.crpix,
        fw.sip.a_order,
        fw.sip.b_order,
        a_coeff,
        b_coeff
    ) + Identity(2)

    unit_conv = Scale(1.0 / 3600.0, name='arcsec_to_deg_1D')
    unit_conv = unit_conv & unit_conv
    unit_conv.name = 'arcsec_to_deg_2D'

    unit_conv_inv = Scale(3600.0, name='deg_to_arcsec_1D')
    unit_conv_inv = unit_conv_inv & unit_conv_inv
    unit_conv_inv.name = 'deg_to_arcsec_2D'

    c2s = CartesianToSpherical(name='c2s', wrap_lon_at=180)
    s2c = SphericalToCartesian(name='s2c', wrap_lon_at=180)
    c2tan = ((Mapping((0, 1, 2), name='xyz') /
              Mapping((0, 0, 0), n_inputs=3, name='xxx')) |
             Mapping((1, 2), name='xtyt'))
    c2tan.name = 'Cartesian 3D to TAN'

    tan2c = (Mapping((0, 0, 1), n_inputs=2, name='xtyt2xyz') |
             (Const1D(1, name='one') & Identity(2, name='I(2D)')))
    tan2c.name = 'TAN to cartesian 3D'

    tan2c.inverse = c2tan
    c2tan.inverse = tan2c

    aff = AffineTransformation2D(matrix=fw.wcs.cd)

    offx = Shift(-fw.wcs.crpix[0])
    offy = Shift(-fw.wcs.crpix[1])

    s = 5e-6
    scale = Scale(s) & Scale(s)

    distortion |= (offx & offy) | scale | tan2c | c2s | unit_conv_inv

    taninv = s2c | c2tan
    tan = Pix2Sky_TAN()
    n2c = RotateNative2Celestial(fw.wcs.crval[0], fw.wcs.crval[1], 180)
    wcslin = unit_conv | taninv | scale.inverse | aff | tan | n2c

    sky_frm = cf.CelestialFrame(reference_frame=coord.ICRS())
    det_frm = cf.Frame2D(name='detector')
    v2v3_frm = cf.Frame2D(
        name="v2v3",
        unit=(u.arcsec, u.arcsec),
        axes_names=('x', 'y'),
        axes_order=(0, 1)
    )
    pipeline = [(det_frm, distortion), (v2v3_frm, wcslin), (sky_frm, None)]

    gw = gwcs.WCS(input_frame=det_frm, output_frame=sky_frm,
                  forward_transform=pipeline)
    gw.crpix = fw.wcs.crpix
    gw.crval = fw.wcs.crval

    # sanity check:
    for _ in range(100):
        x = np.random.randint(1, fw.pixel_shape[0])
        y = np.random.randint(1, fw.pixel_shape[0])
        assert np.allclose(gw(x, y), fw.all_pix2world(x, y, 1),
                           rtol=0, atol=1e-11)

    return gw


def _match(x, y):
    lenx = len(x)
    leny = len(y)
    if lenx == leny:
        return (np.arange(lenx), np.arange(leny))
    elif lenx < leny:
        lenx, leny = leny, lenx
        x, y = y, x
    match = (np.arange(leny) + (0 if y.meta['name'] == 'ext1' else leny),
             np.arange(leny))
    return match


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_multichip_jwst_alignment():
    w1 = _make_gwcs_wcs('data/wfc3_uvis1.hdr')

    imcat1 = tweakwcs.JWSTgWCS(w1, {'v2_ref': 0, 'v3_ref': 0, 'roll_ref': 0})
    imcat1.meta['catalog'] = table.Table.read(
        get_pkg_data_filename('data/wfc3_uvis1.cat'),
        format='ascii.csv',
        delimiter=' ',
        names=['x', 'y']
    )
    imcat1.meta['catalog']['x'] += 1
    imcat1.meta['catalog']['y'] += 1
    imcat1.meta['group_id'] = 1
    imcat1.meta['name'] = 'ext1'

    w2 = _make_gwcs_wcs('data/wfc3_uvis2.hdr')
    imcat2 = tweakwcs.JWSTgWCS(w2, {'v2_ref': 0, 'v3_ref': 0, 'roll_ref': 0})
    imcat2.meta['catalog'] = table.Table.read(
        get_pkg_data_filename('data/wfc3_uvis2.cat'),
        format='ascii.csv',
        delimiter=' ',
        names=['x', 'y']
    )
    imcat2.meta['catalog']['x'] += 1
    imcat2.meta['catalog']['y'] += 1
    imcat2.meta['group_id'] = 1
    imcat2.meta['name'] = 'ext4'

    refcat = table.Table.read(
        get_pkg_data_filename('data/ref.cat'),
        format='ascii.csv', delimiter=' ',
        names=['RA', 'DEC']
    )

    tweakwcs.align_wcs([imcat1, imcat2], refcat, match=_match, nclip=None,
                       sigma=3, fitgeom='general')

    fi1 = imcat1.meta['fit_info']
    fi2 = imcat2.meta['fit_info']

    w1m = imcat1.wcs
    w2m = imcat2.wcs

    assert np.allclose(w1m(*w1.crpix), (83.206917667519, -67.73275818507248), rtol=0)
    assert np.allclose(w2m(*w2.crpix), (83.15167050722597, -67.74220306069903), rtol=0)

    assert np.allclose(fi1['<scale>'], 1.0025, rtol=0, atol=2e-8)
    assert np.allclose(fi2['<scale>'], 1.0025, rtol=0, atol=2e-8)

    assert fi1['rmse'] < 5e-5
    assert fi2['rmse'] < 5e-5

    ra1, dec1 = imcat1.wcs(imcat1.meta['catalog']['x'],
                           imcat1.meta['catalog']['y'])
    ra2, dec2 = imcat2.wcs(imcat2.meta['catalog']['x'],
                           imcat2.meta['catalog']['y'])
    ra = np.concatenate([ra1, ra2])
    dec = np.concatenate([dec1, dec2])
    rra = refcat['RA']
    rdec = refcat['DEC']
    rmse_ra = np.sqrt(np.mean((ra - rra)**2))
    rmse_dec = np.sqrt(np.mean((dec - rdec)**2))

    assert rmse_ra < 3e-9
    assert rmse_dec < 3e-10



"""
A module containing unit tests for the `wcsutil` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
from itertools import product
import math
import pytest
import numpy as np
from astropy.modeling.models import Shift, Rotation2D
from tweakwcs import linearfit, linalg


_LARGE_SAMPLE_SIZE = 1000

_SMALL_SAMPLE_SIZE = 10

_BAD_DATA_FRACTION = 0.2

_TRANSFORM_SELECTOR = {
    'shift': linearfit.fit_shifts,
    'rshift': linearfit.fit_rshift,
    'rscale': linearfit.fit_rscale,
    'general': linearfit.fit_general,
}

_ATOL = 10 * _LARGE_SAMPLE_SIZE * np.sqrt(
    np.finfo(linalg._MAX_LINALG_TYPE).eps
)


def test_test_transform_selector():
    assert(set(_TRANSFORM_SELECTOR) == set(linearfit.SUPPORTED_FITGEOM_MODES))


@pytest.fixture(scope="module")
def ideal_small_data(request):
    # rscale data with proper rotations and no noise
    uv = np.random.random((_SMALL_SAMPLE_SIZE, 2))
    xy = np.random.random((_SMALL_SAMPLE_SIZE, 2))
    wuv = np.random.random(_SMALL_SAMPLE_SIZE)
    wxy = np.random.random(_SMALL_SAMPLE_SIZE)
    return uv, xy, wuv, wxy


@pytest.fixture(scope="function", params=[
    'shifts', 'rshift', 'rscale', 'rscale-flip-x', 'rscale-flip-y', 'affine'
])
def ideal_large_data(request):
    # rscale data with proper rotations and no noise
    uv = np.random.random((_LARGE_SAMPLE_SIZE, 2))

    # assume an image size of 4096x2048:
    uv[:, 0] *= 2047.0
    uv[:, 1] *= 4095.0

    # rotation angle(s):
    angle = 360.0 * np.random.random()  # 0 ... 360

    if request.param == 'shifts':
        angle = (0, 0)
        scale = (1, 1)
        proper = True
        transform = 'shift'

    elif request.param == 'rshift':
        angle = (angle, angle)
        scale = 1
        proper = True
        transform = 'rshift'

    elif request.param == 'rscale':
        angle = (angle, angle)
        scale = 2 * (0.8 + 0.4 * np.random.random(), )  # 0.8 ... 1.2
        proper = True
        transform = 'rscale'

    elif request.param == 'rscale-flip-x':
        angle = ((angle + 180.0) % 360.0, angle)
        scale = 2 * (0.8 + 0.4 * np.random.random(), )  # 0.8 ... 1.2
        proper = False
        transform = 'rscale'

    elif request.param == 'rscale-flip-y':
        angle = (angle, (angle + 180.0) % 360.0)
        scale = 2 * (0.8 + 0.4 * np.random.random(), )  # 0.8 ... 1.2
        proper = False
        transform = 'rscale'

    elif request.param == 'affine':
        # rotation angles:
        offset = 150.0 * (np.random.random() - 0.5)  # -75 ... 75
        offset += 180.0 * np.random.choice([0.0, 1.0])  # add random axis flip
        angle = (angle, (angle + offset) % 360.0)
        # scales:
        scale = 0.8 + 0.4 * np.random.random(2)  # 0.8 ... 1.2
        # proper:
        rad = np.deg2rad(angle)
        proper = (np.prod(np.cos(rad)) + np.prod(np.sin(rad))) > 0
        transform = 'general'

    shift = 200.0 * (np.random.random(2) - 0.5)  # -100 ... +100
    rmat = linearfit.build_fit_matrix(angle, scale)
    skew = angle[1] - angle[0]

    # apply rscale
    xy = np.dot(uv, rmat.T) + shift

    return uv, xy, angle, scale, shift, rmat, proper, skew, transform


@pytest.fixture(scope="function",
                params=[v for v in product(*(2 * [[False, True]]))])
def weight_data(request):
    nbd = int(_BAD_DATA_FRACTION * _LARGE_SAMPLE_SIZE)
    minv = 1000.0
    maxv = 1.0e6

    if not any(request.param):
        wxy = None
        wuv = None
        idx_xy = (np.array([], dtype=int), )
        idx_uv = (np.array([], dtype=int), )
        bd_xy = np.zeros((0, 2))
        bd_uv = np.zeros((0, 2))

    elif all(request.param):
        wxy = np.random.random(_LARGE_SAMPLE_SIZE)
        wuv = np.random.random(_LARGE_SAMPLE_SIZE)

        # split indices into (almost) equal parts:
        nbdxy = int((0.3 + 0.4 * np.random.random()) * nbd)
        nbduv = nbd - nbdxy
        idx_xy = (np.random.choice(np.arange(_LARGE_SAMPLE_SIZE),
                                   nbdxy, replace=False), )
        idx_uv = (np.random.choice(np.arange(_LARGE_SAMPLE_SIZE),
                                   nbduv, replace=False), )

        wxy[idx_xy] = 0.0
        wuv[idx_uv] = 0.0

        sign = np.random.choice([-1, 1], (nbdxy, 2))
        bd_xy = sign * (minv + (maxv - minv) * np.random.random((nbdxy, 2)))

        sign = np.random.choice([-1, 1], (nbduv, 2))
        bd_uv = sign * (minv + (maxv - minv) * np.random.random((nbduv, 2)))

    elif request.param[0] and not request.param[1]:
        idx = np.random.choice(np.arange(_LARGE_SAMPLE_SIZE),
                               nbd, replace=False)
        idx_xy = (idx, )
        idx_uv = (np.array([], dtype=int), )
        wxy = np.random.random(_LARGE_SAMPLE_SIZE)
        wxy[idx_xy] = 0.0
        wuv = None
        sign = np.random.choice([-1, 1], (nbd, 2))
        bd_xy = sign * (minv + (maxv - minv) * np.random.random((nbd, 2)))
        bd_uv = np.zeros((0, 2))

    else:
        idx = np.random.choice(np.arange(_LARGE_SAMPLE_SIZE), nbd,
                               replace=False)
        idx_uv = (idx, )
        idx_xy = (np.array([], dtype=int), )
        wuv = np.random.random(_LARGE_SAMPLE_SIZE)
        wuv[idx_uv] = 0.0
        wxy = None
        sign = np.random.choice([-1, 1], (nbd, 2))
        bd_uv = sign * (minv + (maxv - minv) * np.random.random((nbd, 2)))
        bd_xy = np.zeros((0, 2))

    return wxy, wuv, idx_xy, idx_uv, bd_xy, bd_uv


@pytest.fixture(scope="module")
def tiny_zero_data():
    """ Return a tuple of (xy, uv, wxy, wuv)"""
    return np.zeros((3, 2)), np.zeros((3, 2)), np.zeros(3), np.zeros(3)


@pytest.fixture(scope="function", params=[
    linearfit.fit_shifts, linearfit.fit_rscale, linearfit.fit_general
])
def fit_functions(request):
    return request.param


def test_build_fit_matrix_identity():
    i = np.identity(2)

    r = linearfit.build_fit_matrix(0)  # also test that default scale value = 1
    assert np.allclose(i, r, rtol=0, atol=_ATOL)

    r = linearfit.build_fit_matrix((0, 0), (1, 1))
    assert np.allclose(i, r, rtol=0, atol=_ATOL)


@pytest.mark.parametrize('rot', [1, 35, 75, 95, 155, 189, 261, 299, 358])
def test_build_fit_matrix_rot(rot):
    i = np.identity(2)
    m = linearfit.build_fit_matrix(rot)

    minv = linearfit.build_fit_matrix(360 - rot)
    assert np.allclose(i, np.dot(m, minv), rtol=0, atol=_ATOL)


@pytest.mark.parametrize('rot, scale', [
    ((1, 4), (2.4, 5.6)),
    ((31, 78), (0.9, 1.3)),
])
def test_build_fit_matrix_generalized(rot, scale):
    i = np.identity(2)
    m = linearfit.build_fit_matrix(rot, scale)

    # check scale:
    assert np.allclose(np.sqrt(np.sum(m**2, axis=0)), scale,
                       rtol=0, atol=_ATOL)
    ms = np.diag(scale)

    # check rotations:
    mr = linearfit.build_fit_matrix(rot, 1)
    mrinv = linearfit.build_fit_matrix(rot[::-1], 1).T
    assert np.allclose(np.linalg.det(mr) * i, np.dot(mr, mrinv),
                       rtol=0, atol=_ATOL)

    assert np.allclose(m, np.dot(mr, ms), rtol=0, atol=_ATOL)


@pytest.mark.parametrize('uv, xy, wuv, wxy', [
    (np.zeros(10), np.zeros(10), None, None),
    (np.zeros((10, 2, 2)), np.zeros(10), None, None),
    (np.zeros((10, 2)), np.zeros((11, 2)), None, None),
    3 * (np.zeros((10, 2)), ) + (None, ),
    2 * (np.zeros((10, 2)), ) + (None, np.zeros((10, 2))),
    2 * (np.zeros((10, 2)), ) + (None, np.zeros((5, 2))),
    2 * (np.zeros((10, 2)), ) + (np.zeros((5, 2)), None),
])
def test_iter_linear_fit_invalid_shapes(uv, xy, wuv, wxy):
    # incorrect coordinate array dimensionality:
    with pytest.raises(ValueError):
        linearfit.iter_linear_fit(xy, uv, wxy=wxy, wuv=wuv)


@pytest.mark.parametrize('nclip, sigma', [
    (3, None), (-3, None), (3, -1), (-1, 3), (3, (1.0, 'invalid')),
])
def test_iter_linear_fit_invalid_sigma_nclip(ideal_small_data, nclip, sigma):
    uv, xy, _, _ = ideal_small_data

    with pytest.raises(ValueError):
        linearfit.iter_linear_fit(xy, uv, nclip=nclip, sigma=sigma)


def test_iter_linear_fit_invalid_fitgeom(ideal_small_data):
    uv, xy, _, _ = ideal_small_data

    with pytest.raises(ValueError):
        linearfit.iter_linear_fit(xy, uv, fitgeom='invalid')


@pytest.mark.parametrize('nclip, sigma, clip_accum, weights, noise', [
    (None, 2, True, False, False),
    (None, 2, True, True, False),
    (2, 0.05, False, True, True),
])
def test_iter_linear_fit_special_cases(ideal_large_data, nclip, sigma,
                                       clip_accum, weights, noise):
    uv, xy, _, _, shift, rmat, _, _, fitgeom = ideal_large_data
    if weights:
        wxy, wuv = 0.1 + 0.9 * np.random.random((2, xy.shape[0]))
    else:
        wxy = None
        wuv = None

    if noise:
        xy = xy + np.random.normal(0, 0.01, xy.shape)
        atol = 0.01
    else:
        atol = _ATOL

    fit = linearfit.iter_linear_fit(xy, uv, wxy, wuv, fitgeom=fitgeom,
                                    nclip=nclip, center=(0, 0), sigma=1,
                                    clip_accum=clip_accum)

    assert np.allclose(fit['shift'], shift, rtol=0, atol=atol)
    assert np.allclose(fit['matrix'], rmat, rtol=0, atol=atol)


@pytest.mark.parametrize('weights', [False, True])
def test_iter_linear_fit_1point(weights):
    xy = np.array([[1.0, 2.0]])
    shifts = 20 * (np.random.random(2) - 0.5)
    if weights:
        wxy, wuv = 0.1 + 0.9 * np.random.random((2, xy.shape[0]))
    else:
        wxy, wuv = None, None

    fit = linearfit.iter_linear_fit(xy, xy + shifts, wxy=wxy, wuv=wuv,
                                    fitgeom='shift', nclip=0)

    assert np.allclose(fit['shift'], -shifts, rtol=0, atol=_ATOL)
    assert np.allclose(fit['matrix'], np.identity(2), rtol=0, atol=_ATOL)


def test_iter_linear_fit_fitgeom_clip_all_data(ideal_large_data):
    # Test that clipping is interrupted if number of sources after clipping
    # is below minobj for a given fit:
    xy, uv, _, _, _, _, _, _, fitgeom = ideal_large_data
    ndata = xy.shape[0]
    uv = uv + np.random.normal(0, 0.01, (ndata, 2))
    wxy, wuv = 0.1 + 0.9 * np.random.random((2, ndata))

    fit = linearfit.iter_linear_fit(
        xy, uv, wxy, wuv, fitgeom=fitgeom, sigma=1e-50, nclip=100
    )

    assert np.count_nonzero(fit['fitmask']) == len(xy)
    assert fit['eff_nclip'] == 0


def test_compute_stat_invalid_weights(ideal_small_data):
    pts, _, _, _ = ideal_small_data
    weights = np.zeros(pts.shape[0])
    fit = {}
    linearfit._compute_stat(fit, pts, weights)
    assert math.isnan(fit['rmse'])
    assert math.isnan(fit['mae'])
    assert math.isnan(fit['std'])


@pytest.mark.parametrize('fit_function', [
    linearfit.fit_rscale, linearfit.fit_general,
])
def test_fit_detect_colinear_points(fit_function, tiny_zero_data):
    xy, uv, _, _ = tiny_zero_data
    xy = xy + [1, 2]
    with pytest.raises(linearfit.SingularMatrixError):
        fit_function(xy, uv)


def test_fit_detect_zero_weights(fit_functions, tiny_zero_data):
    xy, uv, wxy, _ = tiny_zero_data
    # all weights are zero:
    with pytest.raises(ValueError):
        fit_functions(xy, uv, wxy=wxy)


def test_fit_detect_negative_weights(fit_functions, tiny_zero_data):
    xy, uv, wuv, _ = tiny_zero_data
    wuv.copy()
    wuv[0] = -1

    # some weights are negative (=invalid):
    with pytest.raises(ValueError):
        fit_functions(xy, uv, wuv=wuv)


@pytest.mark.parametrize('fit_function, npts', [
    (linearfit.fit_shifts, 0),
    (linearfit.fit_rscale, 1),
    (linearfit.fit_general, 2),
])
def test_fit_general_too_few_points(fit_function, npts):
    with pytest.raises(linearfit.NotEnoughPointsError):
        fit_function(np.zeros((npts, 2)), np.zeros((npts, 2)))


@pytest.mark.parametrize(
    'clip_accum, noise',
    [v for v in product(*(2 * [[False, True]]))]
)
def test_iter_linear_fit_clip_style(ideal_large_data, weight_data,
                                    clip_accum, noise):
    """ Test clipping behavior. Test that weights exclude "bad" data. """
    uv, xy, angle, scale, shift, rmat, proper, skew, fitgeom = ideal_large_data
    wxy, wuv, idx_xy, idx_uv, bd_xy, bd_uv = weight_data

    noise_sigma = 0.01
    npts = xy.shape[0]

    # add noise to data
    if noise:
        xy = xy + np.random.normal(0, noise_sigma, (npts, 2))
        atol = 10 * noise_sigma
        nclip = 3
    else:
        atol = _ATOL
        nclip = 0

    if wxy is not None:
        xy[idx_xy] += bd_xy

    if wuv is not None:
        uv = uv.copy()
        uv[idx_uv] += bd_uv

    fit = linearfit.iter_linear_fit(
        xy, uv, wxy=wxy, wuv=wuv, fitgeom=fitgeom, sigma=2,
        clip_accum=clip_accum, nclip=nclip
    )

    shift_with_center = np.dot(rmat, fit['center']) - fit['center'] + shift

    assert np.allclose(fit['shift'], shift_with_center, rtol=0, atol=atol)
    assert np.allclose(fit['matrix'], rmat, rtol=0, atol=atol)
    assert np.allclose(fit['rmse'], 0, rtol=0, atol=atol)
    assert np.allclose(fit['mae'], 0, rtol=0, atol=atol)
    assert np.allclose(fit['std'], 0, rtol=0, atol=atol)
    assert fit['proper'] == proper
    if nclip:
        assert fit['eff_nclip'] > 0
        assert fit['fitmask'].sum(dtype=int) < npts
    else:
        assert fit['eff_nclip'] == 0
        assert (fit['fitmask'].sum(dtype=int) == npts -
                np.union1d(idx_xy[0], idx_uv[0]).size)


def test_iter_rscale_invalid_scale():
    # incorrect coordinate array dimensionality:
    with pytest.raises(ValueError):
        linearfit.fit_rscale(np.zeros((4, 2)), np.zeros((4, 2)), scale=0)


def test_levmar2x2_multivariate():
    inputs = [np.array([10., 10., 20., 20.]), np.array([10., 20., 20., 10.])]
    outputs = [np.array([8.06101731, 0.98994949, 8.06101731, 15.13208512]),
               np.array([12.16223664, 19.23330445, 26.30437226, 19.23330445])]
    rot = Rotation2D()
    rot.fittable = True
    model = (Shift() & Shift()) | rot
    fitter = linearfit._LevMarLSQFitter2x2()
    finfo = fitter(model, inputs, outputs)
    assert np.allclose(finfo.parameters, np.array([4.3, -7.1, 45.]),
                       rtol=1e-5, atol=1e-5)
